from cirrocumulus.simple_data import SimpleData


class FeatureAggregator:

    def __init__(self, obs_measures=[], var_measures=[], dimensions=[]):
        self.obs_measures = obs_measures
        self.var_measures = var_measures
        self.dimensions = dimensions

    @staticmethod
    def add_to_result(df, result):
        if df is not None:
            measure_df = df.T
            for feature in measure_df:
                values = measure_df[feature]
                feature_result = {'min': float(values.loc['min']),
                                  'max': float(values.loc['max']),
                                  'sum': float(values.loc['sum']),
                                  'mean': float(values.loc['mean'])}
                if 'numExpressed' in values.index:
                    feature_result['numExpressed'] = int(values.loc['numExpressed'])
                result[feature] = feature_result

    def execute(self, df):
        result = {}
        for column in self.dimensions:
            df_counts = df.agg({column: lambda x: x.value_counts(sort=False)})
            dimension_summary = {'categories': df_counts.index,
                                 'counts': df_counts[column]}
            result[column] = dimension_summary
        if len(self.var_measures) > 0:
            FeatureAggregator.add_to_result(SimpleData.X_stats(df, self.var_measures), result)
        if len(self.obs_measures) > 0:
            FeatureAggregator.add_to_result(SimpleData.obs_stats(df, self.obs_measures), result)
        return result



import os

import anndata
import pytest

from cirrocumulus.anndata_dataset import AnndataDataset
from cirrocumulus.dataset_api import DatasetAPI


@pytest.fixture(scope='module', autouse=True)
def test_data():
    return anndata.read('test-data/pbmc3k_no_raw.h5ad')


@pytest.fixture(scope='module', autouse=True, params=['small', 'large'])
def measures(request):
    return ['DSCR3', 'TNFRSF4', 'SUMO3'] if request.param == 'small' else list(
        anndata.read('test-data/pbmc3k_no_raw.h5ad').var.index[0:20])


@pytest.fixture(scope='module', autouse=True)
def by():
    return 'louvain'


@pytest.fixture(scope='module', autouse=True)
def dimensions():
    return ['louvain']


@pytest.fixture(scope='module', autouse=True)
def continuous_obs():
    return ['n_genes', 'percent_mito']


@pytest.fixture(scope='module')
def basis():
    return 'X_umap'


@pytest.fixture(scope='module', params=['r', None])
def h5_dataset_backed(request):
    return request.param


@pytest.fixture(scope='module', params=[True, False])
def h5_dataset_force_sparse(request):
    return request.param


@pytest.fixture(scope='module', autouse=True)
def dataset_api(h5_dataset_backed, h5_dataset_force_sparse):
    dataset_api = DatasetAPI()
    dataset_api.add(AnndataDataset(backed=h5_dataset_backed, force_sparse=h5_dataset_force_sparse, extensions=['h5ad']))
    return dataset_api


@pytest.fixture(scope='module', params=['test-data/pbmc3k_no_raw.h5ad'])
def input_dataset(request):
    meta = {'name': os.path.splitext(os.path.basename(request.param))[0], 'url': request.param, 'id': request.param}
    if request.param.endswith('.json'):
        import json
        with open(request.param, 'rt') as f:
            meta.update(json.load(f))

    return meta



# import pandas.io.json as json
import pandas._libs.json as ujson
from flask import make_response

from .file_system_adapter import get_scheme


def to_json(data, response=200):
    # response = make_response(simplejson.dumps(data, check_circular=True), response)
    # response = make_response(json.dumps(data), response)
    s = ujson.dumps(data, double_precision=2, orient='values')
    # s = nujson.dumps(data, double_precision=1)
    response = make_response(s, response)
    response.headers['Content-Type'] = 'application/json'
    return response


def get_email_domain(email):
    at_index = email.find('@')
    domain = None
    if at_index != -1:
        domain = email[at_index + 1:]
    return domain


def load_dataset_schema(url):
    import os

    import fsspec
    import json

    def get_extension(path):
        name, ext = os.path.splitext(path)
        if ext == '.gz':
            name, ext = os.path.splitext(name)
            if ext == '.json':
                ext = '.json.gz'
        return ext

    extension = get_extension(url)
    json_schema = None
    if extension in ['.json', '.json.gz', '']:
        scheme = get_scheme(url)
        fs = fsspec.filesystem(scheme)
        if extension == '':
            url = os.path.join(url, 'index.json.gz')
            extension = get_extension(url)
        if extension == '.json.gz':
            import gzip
            with gzip.open(fs.open(url)) as f:
                json_schema = json.load(f)
        else:
            with fs.open(url) as f:
                json_schema = json.load(f)
    return json_schema



import gzip

import numpy as np
import pandas as pd
import pandas._libs.json as ujson
import scipy.stats as ss

from .data_processing import get_filter_expr, data_filter_keys, get_type_to_measures
from .diff_exp import fdrcorrection
from .envir import CIRRO_SERVE, CIRRO_MAX_WORKERS


executor = None


def submit_job(database_api, dataset_api, email, dataset, job_name, job_type, params):
    global executor
    if executor is None:
        from concurrent.futures.thread import ThreadPoolExecutor
        import os

        if os.environ.get(CIRRO_SERVE) != 'true':
            max_workers = 1
        else:
            max_workers = int(os.environ.get(CIRRO_MAX_WORKERS, '2'))
        executor = ThreadPoolExecutor(max_workers=max_workers)
    job_id = database_api.create_job(email=email, dataset_id=dataset['id'], job_name=job_name, job_type=job_type,
        params=params)
    # run_job(database_api, dataset_api, email, job_id, job_type, dataset, params)
    executor.submit(run_job, database_api, dataset_api, email, job_id, job_type, dataset, params)
    return job_id


def get_mask(dataset_api, dataset, data_filter):
    measures, dimensions, basis_list = data_filter_keys(data_filter)
    keys = get_type_to_measures(measures)
    keys['obs'] = list(dimensions)
    keys['basis'] = basis_list
    df = dataset_api.read_dataset(keys=keys, dataset=dataset)
    return get_filter_expr(df, data_filter)


def run_job(database_api, dataset_api, email, job_id, job_type, dataset, params):
    database_api.update_job(email=email, job_id=job_id, status='running', result=None)
    schema = dataset_api.schema(dataset)
    var_names = schema['var']
    nfeatures = len(var_names)
    mask1 = get_mask(dataset_api, dataset, params['filter'])
    batch_size = 1000  # TODO
    percent_expressed1 = np.zeros(nfeatures)
    percent_expressed2 = np.zeros(nfeatures)
    scores = np.zeros(nfeatures)
    pvals = np.ones(nfeatures)
    is_sparse = None
    v1_size = mask1.sum()

    if job_type == 'de':
        avg_log2FC = np.zeros(nfeatures)
        mask2 = get_mask(dataset_api, dataset, params['filter2'])
        v2_size = mask2.sum()
        
    elif job_type == 'corr':
        df = dataset_api.read_dataset(keys=dict(X=[params['ref']]), dataset=dataset)
        df = df[mask1]
        ref = df[params['ref']]
        is_sparse = hasattr(ref, 'sparse')
        if is_sparse:
            ref = ref.sparse.to_dense()
        is_pearson = False  # params.get('method', 'pearson') == 'pearson'
        if not is_pearson:
            ref_expressed = (ref != 0)
            ref_not_expressed = ~ref_expressed
    index = 0
    for i in range(0, nfeatures, batch_size):
        start = i
        end = min(nfeatures, start + batch_size)
        features = var_names[start:end]
        df = dataset_api.read_dataset(keys=dict(X=features), dataset=dataset)
        df1 = df[mask1]
        if job_type == 'de':
            df2 = df[mask2]
            for feature in features:
                v1 = df1[feature]
                v2 = df2[feature]

                if is_sparse is None:
                    is_sparse = hasattr(v1, 'sparse')
                if is_sparse:
                    v1 = v1.sparse.to_dense()
                    v2 = v2.sparse.to_dense()
                avg_log2FC[index] = np.log2(np.expm1(v1).mean() + 1) - np.log2(np.expm1(v2).mean() + 1)
                percent_expressed1[index] = 100 * ((v1 != 0).sum() / v1_size)
                percent_expressed2[index] = 100 * ((v2 != 0).sum() / v2_size)
                try:
                    scores[index], pvals[index] = ss.mannwhitneyu(v1, v2, alternative="two-sided", use_continuity=False)
                except ValueError:  # All numbers are identical
                    pass
                index += 1
        elif job_type == 'corr':
            for feature in features:
                v1 = df1[feature]
                if is_sparse:
                    v1 = v1.sparse.to_dense()
                try:
                    if is_pearson:
                        scores[index], pvals[index] = ss.pearsonr(ref, v1)
                    else:
                        #              gene1_exp, gene1_not_exp
                        # gene2_exp
                        # gene2_not_exp

                        a = (v1[ref_expressed] != 0).sum()
                        b = (v1[ref_not_expressed] != 0).sum()
                        c = (v1[ref_expressed] == 0).sum()
                        d = (v1[ref_not_expressed] == 0).sum()
                        scores[index], pvals[index] = ss.fisher_exact(
                            [[a, b], [c, d]])
                except ValueError:  # All numbers are identical
                    pass
                index += 1
        database_api.update_job(email=email, job_id=job_id, status='running {:.0f}%'.format(100 * index / nfeatures),
            result=None)
    pvals = fdrcorrection(pvals)
    if job_type == 'de':
        result_df = pd.DataFrame(
            data={'index': var_names, 'comparison:pvals_adj': pvals, 'comparison:avg_log2FC': avg_log2FC,
                  'comparison:scores': scores, 'comparison:percent_expressed1': percent_expressed1,
                  'comparison:percent_expressed2': percent_expressed2})
        result = dict(groups=['comparison'],
            fields=['pvals_adj', 'avg_log2FC', 'scores', 'percent_expressed1', 'percent_expressed2'],
            data=result_df.to_dict(orient='records'))
    elif job_type == 'corr':
        result_df = pd.DataFrame(
            data={'index': var_names, 'selection:pvals_adj': pvals, 'selection:scores': scores})
        result = dict(groups=['selection'], fields=['pvals_adj', 'scores'], data=result_df.to_dict(orient='records'))
    result = ujson.dumps(result, double_precision=2, orient='values').encode('UTF-8')
    result = gzip.compress(result)
    database_api.update_job(email=email, job_id=job_id, status='complete', result=result)



import anndata
import numpy as np
import pandas as pd
import scipy.sparse

from cirrocumulus.abstract_dataset import AbstractDataset
# only works with local files
from cirrocumulus.io_util import read_star_fusion_file
from cirrocumulus.simple_data import SimpleData


class AnndataDataset(AbstractDataset):

    def __init__(self, backed=None, force_sparse=True, extensions=['h5ad', 'loom', 'zarr', 'rds']):
        super().__init__()
        self.path_to_data = {}
        self.backed = backed
        self.force_sparse = force_sparse
        self.extensions = extensions

    def get_suffixes(self):
        return self.extensions

    def read_adata(self, path):
        path_lc = path.lower()
        if path_lc.endswith('.loom'):
            return anndata.read_loom(path)
        elif path_lc.endswith('.zarr'):
            return anndata.read_zarr(path)
        elif path_lc.endswith('.tsv'):
            return read_star_fusion_file(path)
        elif path_lc.endswith('.rds'):  # Seurat, convert to h5ad
            h5_file = path + '.h5ad'
            import os
            if not os.path.exists(h5_file) or abs(os.path.getmtime(h5_file) - os.path.getmtime(path)) > 0.00001:
                import subprocess
                import pkg_resources
                import shutil
                print('Converting Seurat object')
                if os.path.exists(h5_file):
                    os.remove(h5_file)
                subprocess.check_call(
                    ['Rscript', pkg_resources.resource_filename("cirrocumulus", 'seurat2h5ad.R'), path, h5_file])
                shutil.copystat(path, h5_file)
            adata = anndata.read(h5_file, backed=self.backed)
            if adata.raw is not None and adata.shape[0] == adata.raw.shape[0]:
                print('Using adata.raw')
                adata = anndata.AnnData(X=adata.raw.X, var=adata.raw.var, obs=adata.obs, obsm=adata.obsm, uns=adata.uns)
            return adata
        return anndata.read(path, backed=self.backed)
        # elif path.endswith('.mtx'):
        #
        #     return anndata.read_mtx(path, backed=self.backed)
        # elif path.endswith('.txt'):
        #
        #     return anndata.read_text(path, backed=self.backed)
        # elif path.endswith('.csv'):
        #
        #     return anndata.read_csv(path, backed=self.backed)

    def add_data(self, path, data):
        self.path_to_data[path] = data

    def get_data(self, path):
        adata = self.path_to_data.get(path)
        if adata is None:
            adata = self.read_adata(path)
            if scipy.sparse.isspmatrix_csr(adata.X) and adata.X.shape[1] > 1:
                adata.X = adata.X.tocsc()
            self.add_data(path, adata)
        return adata

    def schema(self, file_system, path):
        return SimpleData.schema(self.get_data(path))

    def read_dataset(self, file_system, path, keys=None, dataset=None, schema=None):
        adata = self.get_data(path)
        force_sparse = self.force_sparse
        df = None
        if keys is None:
            keys = {}
        keys = keys.copy()
        var_keys = keys.pop('X', [])
        obs_keys = keys.pop('obs', [])
        basis = keys.pop('basis', [])

        if len(var_keys) > 0:
            X = adata[:, var_keys].X
            if len(X.shape) == 1:
                X = np.array([X]).T
            if force_sparse and not scipy.sparse.issparse(X):
                X = scipy.sparse.csr_matrix(X)
            if scipy.sparse.issparse(X):
                df = pd.DataFrame.sparse.from_spmatrix(X, columns=var_keys)
            else:
                df = pd.DataFrame(data=X, columns=var_keys)
        for key in keys.keys():
            if df is None:
                df = pd.DataFrame()
            d = adata.uns[key]
            features = keys[key]
            X = d['X'][:, d['var'].index.get_indexer_for(features)]
            for i in range(len(features)):
                df[features[i]] = X[:, i]
        if len(obs_keys) > 0:
            if df is None:
                df = pd.DataFrame()
            for key in obs_keys:
                if key == 'index':
                    values = adata.obs.index.values
                else:
                    values = adata.obs[key].values
                df[key] = values

        if basis is not None and len(basis) > 0:
            if df is None:
                df = pd.DataFrame()
            for b in basis:
                embedding_name = b['name']
                embedding_data = adata.obsm[embedding_name]
                dimensions = b['dimensions']
                for i in range(dimensions):
                    df[b['coordinate_columns'][i]] = embedding_data[:, i]
        if df is None:
            df = pd.DataFrame(index=adata.obs.index)
        return df



CIRRO_AUTH_CLIENT_ID = 'CIRRO_AUTH_CLIENT_ID'
CIRRO_DB_URI = 'CIRRO_DB_URI'
CIRRO_DATABASE = 'CIRRO_DATABASE'
CIRRO_EMAIL = 'CIRRO_EMAIL'
CIRRO_SERVE = 'CIRRO_SERVE'
CIRRO_MAX_WORKERS = 'CIRRO_MAX_WORKERS'
CIRRO_FOOTER = 'CIRRO_FOOTER'
CIRRO_BRAND = 'CIRRO_BRAND'
CIRRO_UPLOAD = 'CIRRO_UPLOAD'



import numpy as np


def mean_agg(x):
    return x.mean()


def sparse_sum_agg(x):
    return x.sparse.fill_value if x.sparse.npoints == 0 else x.sum()


def sparse_max_agg(x):
    return x.sparse.fill_value if x.sparse.npoints == 0 else np.max(x.values.sp_values)


def purity_agg(x):
    value_counts = x.value_counts(sort=False)
    largest = value_counts.nlargest(1)
    return largest[0] / value_counts.sum()


def mode_agg(x):
    return x.mode()[0]


def get_basis(basis, nbins=None, agg=None, dimensions=2, precomputed=False, coords=True):
    if isinstance(dimensions, str):
        dimensions = int(dimensions)
    coordinate_columns = []
    for i in range(dimensions):
        coordinate_columns.append(basis + '_' + str(i + 1))
    full_name = basis + '_' + str(dimensions)
    if nbins is not None:
        full_name = full_name + '_' + str(nbins) + '_' + str(agg)
    return {'name': basis, 'dimensions': dimensions, 'coordinate_columns': coordinate_columns, 'nbins': nbins,
            'agg': agg, 'full_name': full_name, 'precomputed': precomputed, 'coords': coords}


class EmbeddingAggregator:

    def __init__(self, measures, dimensions, nbins, basis, agg_function='max', quick=True,
                 coords=True):
        self.nbins = nbins
        self.agg_function = agg_function
        self.measures = measures
        self.dimensions = dimensions
        self.add_count = nbins is not None and len(measures) == 0 and len(dimensions) == 0
        self.basis = basis
        self.quick = quick  # do not compute purity.
        self.coords = coords  # return coordinates

    @staticmethod
    def convert_coords_to_bin(df, nbins, coordinate_columns, bin_name, coordinate_column_to_range=None):
        # replace coordinates with bin, set df[name] to bin
        for name in coordinate_columns:
            values = df[name].values
            if coordinate_column_to_range is not None:
                view_column_range = coordinate_column_to_range[name]
                column_min = view_column_range[0]
                column_max = view_column_range[1]
            else:
                column_min = values.min()
                column_max = values.max()
            df[name] = np.floor(np.interp(values, [column_min, column_max], [0, nbins - 1])).astype(int)
        if len(coordinate_columns) == 2:
            df[bin_name] = df[coordinate_columns[0]] * nbins + df[coordinate_columns[1]]
        else:
            df[bin_name] = df[coordinate_columns[2]] + nbins * (
                    df[coordinate_columns[1]] + nbins * df[coordinate_columns[0]])

    def execute(self, df):
        result = {'coordinates': {}, 'values': {}}
        measures = self.measures
        dimensions = self.dimensions
        add_count = self.add_count
        basis = self.basis
        if basis is not None:
            result['name'] = basis['full_name']
        nbins = self.nbins
        agg_function = self.agg_function
        compute_purity = not self.quick and nbins is not None
        if nbins is not None:
            if basis['full_name'] not in df:
                EmbeddingAggregator.convert_coords_to_bin(df=df, nbins=nbins,
                    coordinate_columns=basis['coordinate_columns'], bin_name=basis['full_name'])

            if add_count:
                df['__count'] = 1.0

            # bin level summary, coordinates have already been converted
            agg_dict = {}
            full_basis_name = basis['full_name']
            for column in basis['coordinate_columns']:
                agg_dict[column] = 'min'
            for column in measures:
                if hasattr(df[column], 'sparse'):
                    if agg_function == 'max':
                        sparse_agg = sparse_max_agg
                    elif agg_function == 'mean':
                        sparse_agg = mean_agg
                    elif agg_function == 'sum':
                        sparse_agg = sparse_sum_agg
                    agg_dict[column] = sparse_agg
                else:
                    agg_dict[column] = agg_function

            for column in dimensions:
                agg_dict[column] = (mode_agg, purity_agg) if compute_purity else mode_agg
            if add_count:
                agg_dict['__count'] = 'sum'

            agg_df = df.groupby(full_basis_name).agg(agg_dict)

            if add_count:
                series = agg_df['__count']
                if compute_purity:
                    series = series[series.columns[0]]
                result['values']['__count'] = series
            if self.coords:
                result['coordinates']['bins'] = agg_df.index

        else:  # no binning
            if add_count:
                result['values']['__count'] = np.ones(len(df))
            agg_df = df
        for column in dimensions:
            series = agg_df[column]
            if nbins is not None:
                if compute_purity:
                    result['values'][column] = dict(value=series['mode_agg'], purity=series['purity_agg'])
                else:
                    result['values'][column] = dict(value=series)
            else:
                result['values'][column] = series
        for column in measures:
            series = agg_df[column]
            if compute_purity:
                series = series[series.columns[0]]
            is_sparse = hasattr(series, 'sparse')
            if is_sparse:
                result['values'][column] = series.sparse.to_dense()
                # result['values'][column] = dict(index=series.values.sp_index, values=series.values.sp_values)
            else:
                result['values'][column] = series

        if self.coords:
            for column in basis['coordinate_columns']:
                series = agg_df[column]
                if compute_purity:
                    series = series[series.columns[0]]
                result['coordinates'][column] = series
        return result



import numpy as np

from cirrocumulus.data_processing import handle_selection_ids


def test_dimension_filter(dataset_api, input_dataset, test_data):
    process_results = handle_selection_ids(dataset_api=dataset_api, dataset=input_dataset,
        data_filter={'filters': [['obs/louvain', 'in', ['1', '5']]]})
    ids_summary = process_results['ids']
    matched_data = test_data[test_data.obs['louvain'].isin(['5', '1'])]
    assert len(np.intersect1d(ids_summary, matched_data.obs.index)) == matched_data.shape[0], '{}, {}'.format(
        len(ids_summary), matched_data.shape[0])


def test_measure_filter(dataset_api, input_dataset, test_data):
    process_results = handle_selection_ids(dataset_api=dataset_api, dataset=input_dataset,
        data_filter={'filters': [['DSCR3', '>', 2]]})
    ids_summary = process_results['ids']
    matched_data = test_data[test_data[:, 'DSCR3'].X > 2]
    assert len(np.intersect1d(ids_summary, matched_data.obs.index)) == matched_data.shape[0]



import numpy as np
import pandas as pd
import scipy.sparse

from cirrocumulus.io_util import cirro_id


class SimpleData:

    def __init__(self, X, obs, var):
        self.obs = obs
        self.var = var
        self.uns = {}
        if X is not None:
            if len(X.shape) == 1:
                X = np.array([X]).T
            n_var = X.shape[1]
            n_obs = X.shape[0]
        else:
            n_var = len(var) if var is not None else 0
            n_obs = len(obs) if obs is not None else 0
        self.X = X
        self.shape = (n_obs, n_var)

    @staticmethod
    def view(adata, row_slice):
        X = adata.X[row_slice] if adata.X is not None else None
        obs = adata.obs[row_slice] if adata.obs is not None else None
        return SimpleData(X, obs, adata.var)

    @staticmethod
    def obs_stats(df, columns):
        df = df[columns]
        # variables on columns, stats on rows, transpose so that stats are on columns
        return df.agg(['min', 'max', 'sum', 'mean']).T

    @staticmethod
    def X_stats(df, var_ids):
        df = df[var_ids]
        if len(df) == 0:
            zeros = np.full(len(var_ids), 0)
            empty = np.full(len(var_ids), np.nan)
            return pd.DataFrame(
                data={'min': empty, 'max': empty, 'sum': zeros,
                      'numExpressed': zeros,
                      'mean': empty}, index=var_ids)

        return pd.DataFrame(
            data={'min': np.min(df.values, axis=0), 'max': np.max(df.values, axis=0), 'sum': df.sum().values,
                  'numExpressed': (df.values != 0).sum(axis=0),
                  'mean': df.mean().values}, index=var_ids)

    @staticmethod
    def get_var_indices(adata, names):
        return adata.var.index.get_indexer_for(names)

    @staticmethod
    def find_markers(adata, key, n_genes):
        import scipy.stats as ss
        import logging
        logger = logging.getLogger("cirro")

        marker_results = []  # array of category, name, id, features
        category = key

        for cat in adata.obs[key].cat.categories:
            logger.info('Computing markers for {}, {}'.format(key, cat))
            mask = adata.obs[key] == cat
            ds1 = adata[mask]
            ds_rest = adata[~mask]
            gene_names_keep = ds1.X.mean(axis=0) > ds_rest.X.mean(axis=0)
            if isinstance(gene_names_keep, np.matrix):
                gene_names_keep = gene_names_keep.A1
            if len(gene_names_keep) == 0:
                continue
            ds1 = ds1[:, gene_names_keep]
            ds_rest = ds_rest[:, gene_names_keep]
            pvals = np.full(ds1.shape[1], 1.0)
            ds1_X = ds1.X
            ds_rest_X = ds_rest.X
            is_sparse = scipy.sparse.isspmatrix(ds1_X)
            for i in range(ds1.shape[1]):
                v1 = ds1_X[:, i]
                v2 = ds_rest_X[:, i]
                if is_sparse:
                    v1 = v1.toarray()[:, 0]
                    v2 = v2.toarray()[:, 0]
                try:
                    _, pvals[i] = ss.mannwhitneyu(v1, v2, alternative="two-sided")
                except ValueError:
                    # All numbers are identical
                    pass
            fc = ds1_X.mean(axis=0) - ds_rest_X.mean(axis=0)
            if isinstance(fc, np.matrix):
                fc = fc.A1
            df = pd.DataFrame(data=dict(pvals=pvals, fc=fc), index=ds1.var_names)
            df = df.sort_values(by=['pvals', 'fc'], ascending=[True, False])
            features = df[:n_genes].index.values
            marker_results.append(dict(category=category, name=str(cat), features=features))
        return marker_results

    @staticmethod
    def schema(adata):
        obs_cat = []
        obs = []
        schema_dict = {'version': '1.0.0'}
        marker_results = []
        prior_marker_results = adata.uns.get('markers', [])
        de_results_format = 'records'
        if isinstance(prior_marker_results, str):
            import json
            prior_marker_results = json.loads(prior_marker_results)
        marker_results += prior_marker_results
        schema_dict['markers'] = marker_results
        n_genes = 10
        scanpy_marker_keys = []

        for key in adata.uns.keys():
            rank_genes_groups = adata.uns[key]
            if isinstance(rank_genes_groups, dict) and 'names' in rank_genes_groups and (
                    'pvals' in rank_genes_groups or 'pvals_adj' in rank_genes_groups or 'scores' in rank_genes_groups):
                scanpy_marker_keys.append(key)
        de_results = []  # array of dicts containing params logfoldchanges, pvals_adj, scores, names

        for scanpy_marker_key in scanpy_marker_keys:
            rank_genes_groups = adata.uns[scanpy_marker_key]
            params = rank_genes_groups['params']
            # pts and pts_rest in later scanpy versions

            rank_genes_groups_keys = list(rank_genes_groups.keys())
            for k in ['params', 'names']:
                if k in rank_genes_groups_keys:
                    rank_genes_groups_keys.remove(k)
            if 'pvals' in rank_genes_groups_keys and 'pvals_adj' in rank_genes_groups_keys:
                rank_genes_groups_keys.remove('pvals')
            category = '{} ({})'.format(params['groupby'], scanpy_marker_key)
            de_result_df = None
            group_names = rank_genes_groups['names'].dtype.names
            for group_name in group_names:
                group_df = pd.DataFrame(index=rank_genes_groups['names'][group_name])

                for rank_genes_groups_key in rank_genes_groups_keys:
                    values = rank_genes_groups[rank_genes_groups_key][group_name]
                    column_name = '{}:{}'.format(group_name, rank_genes_groups_key)
                    group_df[column_name] = values
                group_df = group_df[group_df.index != 'nan']
                if de_result_df is None:
                    de_result_df = group_df
                else:
                    de_result_df = de_result_df.join(group_df, how='outer')
                if n_genes > 0:
                    marker_results.append(
                        dict(category=category, name=str(group_name), features=group_df.index[:n_genes]))

            if de_results_format == 'records':
                de_result_data = de_result_df.reset_index().to_dict(orient='records')
            else:
                de_result_data = dict(index=de_result_df.index)
                for c in de_result_df:
                    de_result_data[c] = de_result_df[c]

            de_result = dict(id=cirro_id(),
                color='logfoldchanges' if 'logfoldchanges' in rank_genes_groups_keys else rank_genes_groups_keys[0],
                size='pvals_adj' if 'pvals_adj' in rank_genes_groups_keys else rank_genes_groups_keys[0],
                params=params, groups=group_names, fields=rank_genes_groups_keys, type='de',
                name=category)
            de_result['data'] = de_result_data
            de_results.append(de_result)

        if 'de_res' in adata.varm:  # pegasus
            de_res = adata.varm['de_res']
            names = de_res.dtype.names
            field_names = set()  # e.g. 1:auroc
            group_names = set()
            for name in names:
                index = name.index(':')
                field_name = name[index + 1:]
                group_name = name[:index]
                field_names.add(field_name)
                group_names.add(group_name)
            group_names = list(group_names)
            field_names = list(field_names)

            de_result_df = pd.DataFrame(data=de_res, index=adata.var.index)
            de_result_df.index.name = 'index'
            if de_results_format == 'records':
                de_result_data = de_result_df.reset_index().to_dict(orient='records')
            else:
                de_result_data = dict(index=de_result_df.index)
                for c in de_res:
                    de_result_data[c] = de_result_df[c]

            de_result = dict(id=cirro_id(), type='de', name='pegasus_de',
                color='log2FC' if 'log2FC' in field_names else field_names[0],
                size='mwu_qval' if 'mwu_qval' in field_names else field_names[0], groups=group_names,
                fields=field_names)
            de_result['data'] = de_result_data
            de_results.append(de_result)

            if n_genes > 0:
                field_use = None
                for field in ['mwu_qval', 'auroc', 't_qval']:
                    if field in field_names:
                        field_use = field
                        break
                if field_use is not None:
                    field_ascending = field_use != 'auroc'
                    for group_name in group_names:
                        fc_column = '{}:log2FC'.format(group_name)
                        name = '{}:{}'.format(group_name, field_name)
                        idx_up = de_result_df[fc_column] > 0
                        df_up = de_result_df.loc[idx_up].sort_values(by=[name, fc_column],
                            ascending=[field_ascending, False])
                        features = df_up[:n_genes].index.values
                        marker_results.append(dict(category='markers', name=str(group_name), features=features))

        for key in adata.obs_keys():
            if pd.api.types.is_categorical_dtype(adata.obs[key]) or pd.api.types.is_bool_dtype(
                    adata.obs[key]) or pd.api.types.is_object_dtype(adata.obs[key]):
                obs_cat.append(key)
            else:
                obs.append(key)
        schema_dict['results'] = de_results
        if 'metagenes' in adata.uns:
            metagenes = adata.uns['metagenes']
            schema_dict['metafeatures'] = metagenes['var'].index

        category_to_order = {}
        for key in adata.obs_keys():
            if pd.api.types.is_categorical_dtype(adata.obs[key]):
                category_to_order[key] = adata.obs[key].cat.categories
        schema_dict['categoryOrder'] = category_to_order
        # spatial_node = adata.uns['spatial'] if 'spatial' in adata.uns else None
        #
        # if spatial_node is not None:
        #     spatial_node_keys = list(spatial_node.keys())  # list of datasets
        #     if len(spatial_node_keys) == 1:
        #         spatial_node = spatial_node[spatial_node_keys[0]]

        images_node = adata.uns.get('images',
            [])  # list of {type:image or meta_image, name:image name, image:path to image, spot_diameter:Number}
        image_names = list(map(lambda x: x['name'], images_node))
        schema_dict['var'] = adata.var_names.values
        schema_dict['obs'] = obs
        schema_dict['obsCat'] = obs_cat
        schema_dict['shape'] = adata.shape

        embeddings = []
        for key in adata.obsm_keys():
            dim = min(3, adata.obsm[key].shape[1])
            if dim < 2:
                continue
            embedding = dict(name=key, dimensions=dim)
            try:
                image_index = image_names.index(key)
                embedding['spatial'] = images_node[image_index]
            except ValueError:
                pass

            if dim == 3:
                embeddings.append(embedding)
                embedding = embedding.copy()
                embedding['dimensions'] = 2
                embeddings.append(embedding)
            else:
                embeddings.append(embedding)
        meta_images = adata.uns.get('meta_images', [])
        for meta_image in meta_images:
            embeddings.append(meta_image)
        schema_dict['embeddings'] = embeddings
        field_to_value_to_color = dict()  # field -> value -> color
        schema_dict['colors'] = field_to_value_to_color

        for key in adata.uns.keys():
            if key.endswith('_colors'):
                field = key[0:len(key) - len('_colors')]
                if field in adata.obs:
                    colors = adata.uns[key]
                    if pd.api.types.is_categorical_dtype(adata.obs[field]):
                        categories = adata.obs[field].cat.categories
                        if len(categories) == len(colors):
                            color_map = dict()
                            for i in range(len(categories)):
                                color_map[str(categories[i])] = colors[i]
                            field_to_value_to_color[field] = color_map
                    else:
                        print("Skipping colors for {}".format(key))
        return schema_dict

    @staticmethod
    def to_df(adata, obs_measures, var_measures, dimensions, basis=None):
        df = pd.DataFrame()
        obs_keys = obs_measures + dimensions
        if basis is not None:
            obs_keys += basis['coordinate_columns']

        for key in obs_keys:
            df[key] = adata.obs[key]
        indices = SimpleData.get_var_indices(adata, var_measures)
        for i in range(len(var_measures)):
            X = adata.X[:, indices[i]]
            if scipy.sparse.issparse(X):
                X = X.toarray()
            df[var_measures[i]] = X
        return df



import os

from .file_system_adapter import FileSystemAdapter


def get_path(dataset, dataset_path):
    path = dataset['url']
    if path[len(path) - 1] == '/':  # remove trailing slash
        path = path[0:len(path) - 1]
    if path.endswith('.json'):
        path = os.path.dirname(path)
    path = path + '/' + dataset_path
    return path


class DatasetAPI:
    def __init__(self):
        self.suffix_to_provider = {}
        self.fs_adapter = FileSystemAdapter()
        self.default_provider = None
        self.cached_schema = None
        self.cached_dataset_id = None

    def get_dataset_provider(self, path):
        index = path.rfind('.')
        if index == -1:
            return self.default_provider
        suffix = path[index + 1:].lower()

        provider = self.suffix_to_provider.get(suffix)
        return provider if provider is not None else self.default_provider

    def add(self, provider):
        if self.default_provider is None:
            self.default_provider = provider
        suffixes = provider.get_suffixes()
        for suffix in suffixes:
            self.suffix_to_provider[suffix.lower()] = provider

    def schema(self, dataset):
        dataset_id = dataset['id']
        if self.cached_dataset_id == dataset_id:
            return self.cached_schema
        path = dataset['url']
        provider = self.get_dataset_provider(path)
        schema_dict = provider.schema(self.fs_adapter.get_fs(path), path)
        if 'summary' in dataset:
            schema_dict['summary'] = dataset['summary']
        if 'markers' in schema_dict:
            schema_dict['markers_read_only'] = schema_dict.pop('markers')
        self.cached_schema = schema_dict
        self.cached_dataset_id = dataset['id']
        return schema_dict

    def has_precomputed_stats(self, dataset):
        path = dataset['url']
        provider = self.get_dataset_provider(path)
        return provider.has_precomputed_stats(self.fs_adapter.get_fs(path), path, dataset)

    def read_precomputed_stats(self, dataset, obs_keys=[], var_keys=[]):
        path = dataset['url']
        provider = self.get_dataset_provider(path)
        return provider.read_precomputed_stats(self.fs_adapter.get_fs(path), path, obs_keys=obs_keys, var_keys=var_keys)

    def read_precomputed_grouped_stats(self, dataset, obs_keys=[], var_keys=[]):
        path = dataset['url']
        provider = self.get_dataset_provider(path)
        return provider.read_precomputed_grouped_stats(self.fs_adapter.get_fs(path), path, obs_keys=obs_keys,
            var_keys=var_keys)

    def read_precomputed_basis(self, dataset, obs_keys=[], var_keys=[], basis=None):
        path = dataset['url']
        provider = self.get_dataset_provider(path)
        return provider.read_precomputed_basis(self.fs_adapter.get_fs(path), path, obs_keys=obs_keys, var_keys=var_keys,
            basis=basis)

    def read_dataset(self, dataset, keys=[]):
        path = dataset['url']
        provider = self.get_dataset_provider(path)
        return provider.read_dataset(self.fs_adapter.get_fs(path), path, keys=keys, dataset=dataset,
            schema=self.schema(dataset))



import fsspec
import pandas as pd
import scipy

from cirrocumulus.embedding_aggregator import get_basis
from cirrocumulus.parquet_dataset import ParquetDataset
from cirrocumulus.prepare_data import PrepareData


def test_prepare(test_data, measures, dimensions, continuous_obs, basis, tmp_path):
    output_dir = str(tmp_path)
    test_data = test_data[:, measures]
    test_data.obs = test_data.obs[dimensions + continuous_obs]
    prepare_data = PrepareData(adata=test_data, output=output_dir)
    prepare_data.execute()
    pq_ds = ParquetDataset()
    dataset = dict(id='')
    schema = dict(shape=test_data.shape)
    fs = fsspec.filesystem('file')
    prepared_df = pq_ds.read_dataset(file_system=fs, path=output_dir, dataset=dataset, schema=schema,
        keys=dict(X=measures, obs=dimensions + continuous_obs, basis=[get_basis(basis, -1, '')]))

    if not scipy.sparse.issparse(test_data.X):
        test_data.X = scipy.sparse.csr_matrix(test_data.X)
    df = pd.DataFrame.sparse.from_spmatrix(test_data.X, columns=measures)

    for f in dimensions:
        df[f] = test_data.obs[f].values
        df[f] = df[f].astype('category')
    for f in continuous_obs:
        df[f] = test_data.obs[f].values
    embedding_data = test_data.obsm[basis]

    for i in range(embedding_data.shape[1]):
        df["{}_{}".format(basis, i + 1)] = embedding_data[:, i]
    prepared_df = prepared_df[df.columns]
    pd.testing.assert_frame_equal(df, prepared_df, check_names=False)



import numpy as np
import pandas as pd
import pytest

from cirrocumulus.data_processing import handle_data
from cirrocumulus.embedding_aggregator import EmbeddingAggregator, get_basis


def create_df(test_data, measures, dimensions, basis):
    df = pd.DataFrame(test_data[:, measures].X.toarray(), columns=measures)
    df = df.join(test_data.obs[dimensions].reset_index())
    return df.join(pd.DataFrame(test_data.obsm['X_umap'][:, 0:2], columns=basis['coordinate_columns']))


def group_df(test_data, measures, dimensions, continuous_obs, basis):
    df = create_df(test_data, measures, dimensions + continuous_obs, basis)
    agg_dict = {}
    for key in basis['coordinate_columns']:
        agg_dict[key] = 'max'
    for key in measures + continuous_obs:
        agg_dict[key] = basis['agg']
    for key in dimensions:
        agg_dict[key] = lambda x: x.mode()[0]
    EmbeddingAggregator.convert_coords_to_bin(df, nbins=100,
        coordinate_columns=basis['coordinate_columns'],
        bin_name=basis['full_name'])
    # group by bin then agg
    return df.groupby(basis['full_name']).agg(agg_dict)


def diff_binning(grouped_df, measures, dimensions, continuous_obs, basis, results):
    for key in basis['coordinate_columns']:
        np.testing.assert_array_equal(results['coordinates'][key], grouped_df[key].values, err_msg=key)
    for key in measures:
        np.testing.assert_allclose(results['values'][key], grouped_df[key].values, err_msg=key, rtol=1e-05, atol=1e-06)
    for key in continuous_obs:
        np.testing.assert_allclose(results['values'][key], grouped_df[key].values, err_msg=key, rtol=1e-05,
            atol=1e-08)
    for key in dimensions:
        np.testing.assert_array_equal(results['values'][key]['value'], grouped_df[key].values, err_msg=key)


def test_no_binning(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs, basis):
    embedding_list = [dict(ndim=2, basis=basis)]
    basis = get_basis(basis)
    values = dict(dimensions=dimensions, measures=measures + list(map(lambda x: 'obs/' + x, continuous_obs)))
    results = handle_data(dataset_api=dataset_api, dataset=input_dataset, values=values, embedding_list=embedding_list)
    df = create_df(test_data, measures, dimensions + continuous_obs, basis)
    for key in basis['coordinate_columns']:
        np.testing.assert_array_equal(results['embeddings'][0]['coordinates'][key], df[key].values, err_msg=key)
    for key in measures:
        np.testing.assert_array_equal(results['values'][key], df[key].values, err_msg=key)
    for key in dimensions:
        np.testing.assert_array_equal(results['values'][key], test_data.obs[key], err_msg=key)


@pytest.fixture(autouse=True, params=['sum', 'mean', 'max'])
def agg_function(request):
    return request.param


def test_binning(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs, basis,
                 agg_function):
    nbins = 100
    basis_dict = get_basis(basis, nbins, agg_function)
    grouped_df = group_df(test_data, measures, dimensions, continuous_obs, basis_dict)
    embedding_list = [dict(nbins=nbins, ndim=2, basis=basis, dimensions=dimensions, agg=agg_function)]
    embedding_list[0]['measures'] = measures + list(map(lambda x: 'obs/' + x, continuous_obs))
    results = handle_data(dataset_api=dataset_api, dataset=input_dataset, embedding_list=embedding_list)
    diff_binning(grouped_df, measures, dimensions, continuous_obs, basis_dict, results['embeddings'][0])



from urllib.parse import urlparse

import fsspec


def get_scheme(path):
    pr = urlparse(path)
    if len(pr.scheme) <= 1:  # for file paths: /foo/bar/test.h5ad or C:/foo/bar/test.h5ad
        return 'file'
    return pr.scheme


class FileSystemAdapter:

    def __init__(self):
        # scheme can be gs, file, etc.
        self.scheme_to_fs = {}

    def get_fs(self, path):
        scheme = get_scheme(path)
        fs = self.scheme_to_fs.get(scheme, None)
        if fs is not None:
            return fs
        fs = fsspec.filesystem(scheme)
        self.scheme_to_fs[scheme] = fs
        return fs



import cachecontrol
import google.auth.transport.requests
import requests
from flask import request
from google.oauth2 import id_token


class GoogleAuth:

    def __init__(self, clientId):
        session = requests.session()
        self.clientId = clientId
        cached_session = cachecontrol.CacheControl(session)
        self.cached_request = google.auth.transport.requests.Request(session=cached_session)

    @property
    def client_id(self):
        return self.clientId

    def auth(self):
        token = request.headers.get('Authorization')
        if token is not None:
            if not token.startswith('Bearer '):
                raise ValueError('Token should start with Bearer ')
            token = token.split('Bearer ')[1]
        else:
            token = request.args.get('access_token')
        idinfo = id_token.verify_oauth2_token(token, self.cached_request, self.clientId)
        if idinfo['aud'] != self.clientId:
            raise ValueError('Wrong aud')
        if idinfo['iss'] not in ['accounts.google.com', 'https://accounts.google.com']:
            raise ValueError('Wrong issuer.')
        return idinfo



import pandas as pd


class DotPlotAggregator:

    def __init__(self, var_measures, dimensions):
        self.var_measures = var_measures
        self.dimensions = dimensions

    def execute(self, df):
        results = []
        # {categories:[], name:'', values:[{name:'', percentExpressed:0, mean:0}]}
        var_measures = self.var_measures
        dimensions = self.dimensions
        if len(var_measures) == 0 or len(dimensions) == 0:
            return results

        def mean(x):
            return x.mean()

        def percent_expressed(x):
            return 100 * ((x.values != 0).sum() / len(x))

        for d in dimensions:
            dimension_name = d
            if isinstance(d, list):
                if len(d) > 1:
                    dimension_name = '-'.join(d)
                    df[dimension_name] = df[d[0]].astype(str).str.cat(df[d[1:]].astype(str), sep="-").astype('category')
                else:
                    dimension_name = d[0]
            if pd.api.types.is_categorical_dtype(df[dimension_name]) and len(df[dimension_name].dtype.categories) <= 1:
                continue
            agg_result = df.groupby(dimension_name, observed=True).agg([mean, percent_expressed])

            values = []
            dotplot_result = {'categories': agg_result.index, 'name': dimension_name, 'values': values}
            for var_measure in var_measures:
                series = agg_result[var_measure]
                is_sparse = hasattr(series, 'sparse')
                if is_sparse:
                    series = series.sparse.to_dense()
                values.append({'name': var_measure,
                               'percentExpressed': series['percent_expressed'],
                               'mean': series['mean']})
            results.append(dotplot_result)
        return results



import os

from flask import Blueprint, Response, request, stream_with_context, current_app

import cirrocumulus.data_processing as data_processing
from .dataset_api import DatasetAPI
from .envir import CIRRO_SERVE, CIRRO_FOOTER, CIRRO_UPLOAD, CIRRO_BRAND
from .file_system_adapter import get_scheme
from .invalid_usage import InvalidUsage
from .job_api import submit_job
from .util import to_json

blueprint = Blueprint('blueprint', __name__)

dataset_api = DatasetAPI()


@blueprint.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    return Response(error.message, error.status_code)


def get_database():
    return current_app.config['DATABASE']


def get_auth():
    return current_app.config['AUTH']


@blueprint.route('/server', methods=['GET'])
def handle_server():
    # no login required
    server = get_database().server()
    server['clientId'] = get_auth().client_id
    if os.environ.get(CIRRO_FOOTER) is not None:
        with open(os.environ.get(CIRRO_FOOTER), 'rt') as f:
            server['footer'] = f.read()
    if os.environ.get(CIRRO_BRAND) is not None:
        with open(os.environ.get(CIRRO_BRAND), 'rt') as f:
            server['brand'] = f.read()
    # server['brand'] = os.environ.get(CIRRO_BRAND)
    server['jobs'] = os.environ.get('GAE_APPLICATION') is None
    server['upload'] = os.environ.get(CIRRO_UPLOAD) is not None
    return to_json(server)


@blueprint.route('/filters', methods=['GET'])
def handle_dataset_filters():
    """List filters available for a dataset.
    """
    email = get_auth().auth()['email']
    dataset_id = request.args.get('id', '')

    if dataset_id == '':
        return 'Please provide an id', 400

    dataset_filters = get_database().dataset_filters(email, dataset_id)
    # {'id': result.id, 'name': result['name']}
    return to_json(dataset_filters)


@blueprint.route('/export_filters', methods=['GET'])
def handle_export_dataset_filters():
    """Download filters in a csv file for a dataset.
    """
    email = get_auth().auth()['email']
    dataset_id = request.args.get('id', '')

    if dataset_id == '':
        return 'Please provide an id', 400
    database_api = get_database()
    dataset_filters = database_api.dataset_filters(email, dataset_id)
    dataset = database_api.get_dataset(email, dataset_id)
    text = data_processing.handle_export_dataset_filters(dataset_api, dataset, dataset_filters)
    return Response(text, mimetype='text/plain')


@blueprint.route('/category_name', methods=['GET', 'PUT'])
def handle_category_name():
    """CRUD for category name.
    """

    email = get_auth().auth()['email']
    database_api = get_database()
    if request.method == 'GET':
        dataset_id = request.args.get('id', '')
        if dataset_id == '':
            return 'Please provide an id', 400
        return to_json(database_api.category_names(email, dataset_id))
    content = request.get_json(force=True, cache=False)
    category = content.get('c')
    original_name = content.get('o')
    new_name = content.get('n')
    dataset_id = content.get('id')
    if request.method == 'PUT':
        database_api.upsert_category_name(
            email=email,
            category=category,
            dataset_id=dataset_id,
            original_name=original_name,
            new_name=new_name)
        return '', 200


@blueprint.route('/feature_set', methods=['POST', 'DELETE'])
def handle_feature_set():
    """CRUD for a feature set.
    """

    email = get_auth().auth()['email']
    database_api = get_database()
    # if request.method == 'GET':
    #     filter_id = request.args.get('id', '')
    #     dataset_id = request.args.get('ds_id', '')
    #     if filter_id == '' or dataset_id == '':
    #         return 'Please provide an id', 400
    #     return to_json(database_api.get_dataset_filter(email, dataset_id=dataset_id, filter_id=filter_id))
    content = request.get_json(force=True, cache=False)
    set_id = content.get('id')
    dataset_id = content.get('ds_id')
    # POST=new, PUT=update , DELETE=delete, GET=get
    if request.method == 'PUT' or request.method == 'POST':
        if request.method == 'PUT' and set_id is None:
            return 'Please supply an id', 400
        if request.method == 'POST' and dataset_id is None:
            return 'Please supply a ds_id', 400

        name = content.get('name')
        category = content.get('category')
        features = content.get('features')
        set_id = database_api.upsert_feature_set(
            email=email,
            dataset_id=dataset_id,
            set_id=set_id if request.method == 'PUT' else None,
            name=name,
            category=category,
            features=features)
        return to_json({'id': set_id})
    elif request.method == 'DELETE':
        database_api.delete_feature_set(email, dataset_id=dataset_id, set_id=set_id)
        return to_json('', 204)


@blueprint.route('/filter', methods=['GET', 'POST', 'PUT', 'DELETE'])
def handle_dataset_filter():
    """CRUD for a dataset filter.
    """

    email = get_auth().auth()['email']
    database_api = get_database()
    if request.method == 'GET':
        filter_id = request.args.get('id', '')
        dataset_id = request.args.get('ds_id', '')
        if filter_id == '' or dataset_id == '':
            return 'Please provide an id', 400
        return to_json(database_api.get_dataset_filter(email, dataset_id=dataset_id, filter_id=filter_id))
    content = request.get_json(force=True, cache=False)
    filter_id = content.get('id')
    dataset_id = content.get('ds_id')
    # POST=new, PUT=update , DELETE=delete, GET=get
    if request.method == 'PUT' or request.method == 'POST':
        if request.method == 'PUT' and filter_id is None:
            return 'Please supply an id', 400
        if request.method == 'POST' and dataset_id is None:
            return 'Please supply a ds_id', 400
        dataset_filter = content.get('value')
        filter_name = content.get('name')
        filter_notes = content.get('notes')
        filter_id = database_api.upsert_dataset_filter(
            email=email,
            dataset_id=dataset_id,
            filter_id=filter_id if request.method == 'PUT' else None,
            dataset_filter=dataset_filter,
            filter_name=filter_name,
            filter_notes=filter_notes)
        return to_json({'id': filter_id})
    elif request.method == 'DELETE':
        database_api.delete_dataset_filter(email, dataset_id=dataset_id, filter_id=filter_id)
        return to_json('', 204)


@blueprint.route('/schema', methods=['GET'])
def handle_schema():
    """Get dataset schema.
    """
    email = get_auth().auth()['email']
    database_api = get_database()
    dataset_id = request.args.get('id', '')
    dataset = database_api.get_dataset(email, dataset_id)
    schema = dataset_api.schema(dataset)
    schema.update(dataset)  # add title, etc from database to schema
    schema['markers'] = database_api.get_feature_sets(email=email, dataset_id=dataset_id)
    return to_json(schema)


def get_file_path(file, dataset_url):
    # when serving dataset, image must be relative to dataset directory
    if os.environ.get(CIRRO_SERVE) == 'true':
        _, ext = os.path.splitext(dataset_url)
        if file[0] == '/' or file.find('..') != -1:
            raise ValueError('Incorrect path')
        file_path = os.path.join(dataset_url, file)
    else:
        file_path = file
        scheme = get_scheme(file_path)
        if scheme == 'file' and not os.path.exists(file_path):
            file_path = os.path.join(dataset_url, file)
    return file_path


def send_file(file_path):
    import mimetypes
    mimetype, encoding = mimetypes.guess_type(file_path)
    # with dataset_api.fs_adapter.get_fs(file_path).open(file_path) as f:
    #     bytes = f.read()
    # return Response(bytes, mimetype=mimetype[0])
    chunk_size = 4096
    f = dataset_api.fs_adapter.get_fs(file_path).open(file_path)

    def generate():
        while True:
            chunk = f.read(chunk_size)
            if len(chunk) <= 0:
                f.close()
                break
            yield chunk

    r = Response(stream_with_context(generate()), mimetype=mimetype)
    r.headers["Content-Encoding"] = encoding
    return r


@blueprint.route('/file', methods=['GET'])
def handle_file():
    email = get_auth().auth()['email']
    database_api = get_database()
    dataset_id = request.args.get('id', '')
    dataset = database_api.get_dataset(email, dataset_id)
    file_path = get_file_path(request.args.get('file'), dataset['url'])
    return send_file(file_path)


@blueprint.route('/user', methods=['GET'])
def handle_user():
    email = get_auth().auth()['email']
    database_api = get_database()
    user = database_api.user(email)
    return to_json(user)


def get_email_and_dataset(content):
    email = get_auth().auth()['email']
    dataset_id = content.get('id', '')
    if dataset_id == '':
        return 'Please supply an id', 400
    database_api = get_database()
    dataset = database_api.get_dataset(email, dataset_id)
    return email, dataset


@blueprint.route('/data', methods=['POST'])
def handle_data():
    json_request = request.get_json(cache=False)
    email, dataset = get_email_and_dataset(json_request)

    return to_json(
        data_processing.handle_data(dataset_api=dataset_api, dataset=dataset,
            embedding_list=json_request.get('embedding'),
            values=json_request.get('values'),
            stats=json_request.get('stats'),
            grouped_stats=json_request.get('groupedStats'),
            selection=json_request.get('selection')))


@blueprint.route('/selection', methods=['POST'])
def handle_selection():
    # selection includes stats, coordinates
    content = request.get_json(cache=False)
    email, dataset = get_email_and_dataset(content)
    data_filter = content.get('filter')
    return to_json(data_processing.handle_selection(dataset_api=dataset_api, dataset=dataset, data_filter=data_filter,
        measures=content.get('measures', []), dimensions=content.get('dimensions', []),
        embeddings=content.get('embeddings', [])))


@blueprint.route('/selected_ids', methods=['POST'])
def handle_selected_ids():
    content = request.get_json(cache=False)
    email, dataset = get_email_and_dataset(content)
    data_filter = content.get('filter')
    return to_json(
        data_processing.handle_selection_ids(dataset_api=dataset_api, dataset=dataset, data_filter=data_filter))


# List available datasets
@blueprint.route('/datasets', methods=['GET'])
def handle_datasets():
    email = get_auth().auth()['email']
    database_api = get_database()
    return to_json(database_api.datasets(email))


def upload_file(file):
    from werkzeug.utils import secure_filename
    filename = secure_filename(file.filename)
    dest = os.path.join(os.environ.get(CIRRO_UPLOAD), filename)
    dest_fs = dataset_api.fs_adapter.get_fs(dest).open(dest, mode='wb')
    file.save(dest_fs)
    dest_fs.close()
    return dest


def copy_url(url):
    import fsspec
    from werkzeug.utils import secure_filename
    upload = os.environ.get(CIRRO_UPLOAD)

    if url.endswith('/'):  # don't copy directories
        return url
    from urllib.parse import urlparse
    if urlparse(upload).netloc == urlparse(url).netloc:  # don't copy if already in the same bucket
        return url
    src_scheme = get_scheme(url)
    src_fs = fsspec.filesystem(src_scheme)
    filename = secure_filename(os.path.basename(url))
    dest = os.path.join(upload, filename)
    dest_scheme = get_scheme(dest)
    dest_fs = fsspec.filesystem(dest_scheme)
    out = dest_fs.open(dest, 'wb')
    n = 1024 * 1024
    with src_fs.open(url, mode='rb') as r:
        while True:
            b = r.read(n)
            if not b:
                break
            out.write(b)
    out.close()
    return dest


@blueprint.route('/dataset', methods=['GET', 'POST', 'PUT', 'DELETE'])
def handle_dataset():
    email = get_auth().auth()['email']
    database_api = get_database()
    # POST=new dataset, PUT=update dataset, DELETE=delete, GET=get dataset info
    if request.method == 'PUT' or request.method == 'POST':
        dataset_id = request.form.get('id')
        dataset_name = request.form.get('name')
        description = request.form.get('description')
        species = request.form.get('species')
        title = request.form.get('title')
        url = request.form.get('url')  # e.g. gs://foo/a/b/
        file = request.files.get('file')
        readers = request.form.get('readers')
        if readers is not None:
            import json
            readers = json.loads(readers)
        if request.method == 'PUT' and dataset_id is None:  # update
            return 'Please supply an id', 400
        if request.method == 'POST' and dataset_name == '':  # new
            return 'Must supply dataset name', 400

        if url is not None and os.environ.get(CIRRO_UPLOAD) is not None:
            url = copy_url(url)

        if file is not None:
            if os.environ.get(CIRRO_UPLOAD) is None:
                return 'Upload not supported', 400
            url = upload_file(file)
        dataset_id = database_api.upsert_dataset(email=email,
            dataset_id=dataset_id if request.method == 'PUT' else None,
            dataset_name=dataset_name, url=url, readers=readers, description=description, title=title, species=species)
        return to_json({'id': dataset_id})
    elif request.method == 'DELETE':
        content = request.get_json(force=True, cache=False)
        dataset_id = content.get('id', '')
        database_api.delete_dataset(email, dataset_id)
        return to_json('', 204)
    elif request.method == 'GET':
        dataset_id = request.args.get('id', '')
        return to_json(database_api.get_dataset(email, dataset_id, True))


@blueprint.route('/job', methods=['GET', 'DELETE'])
def handle_job():
    email = get_auth().auth()['email']
    database_api = get_database()
    if request.method == 'DELETE':
        content = request.get_json(force=True, cache=False)
        job_id = content.get('id', '')
        database_api.delete_job(email, job_id)
        return to_json('', 204)
    else:
        job_id = request.args.get('id', '')
        job_status = request.args.get('status', '0')
        return_result = job_status == '0'
        if job_id.startswith('cirro-'):  # precomputed result
            dataset_id = request.args.get('ds_id', '')
            email = get_auth().auth()['email']
            dataset = database_api.get_dataset(email, dataset_id)
            file_path = get_file_path(os.path.join('uns', job_id + '.json.gz'), dataset['url'])
            return send_file(file_path)
        result = database_api.get_job(email=email, job_id=job_id, return_result=return_result)
        if return_result:
            result = Response(result, mimetype='application/json')
            result.headers["Content-Encoding"] = 'gzip'
        return result


@blueprint.route('/jobs', methods=['GET'])
def handle_jobs():
    email = get_auth().auth()['email']
    database_api = get_database()
    ds_id = request.args.get('id', '')
    return to_json(database_api.get_jobs(email=email, dataset_id=ds_id))


@blueprint.route('/submit_job', methods=['POST'])
def handle_submit_job():
    if os.environ.get('GAE_APPLICATION') is None:  # TODO
        database_api = get_database()
        content = request.get_json(force=True, cache=False)
        email, dataset = get_email_and_dataset(content)
        params = content.get('params')
        job_type = content.get('type')
        job_name = content.get('name')
        return dict(id=submit_job(database_api=database_api, dataset_api=dataset_api, email=email, dataset=dataset,
            job_name=job_name, job_type=job_type, params=params))



import argparse
import gzip
import json
import logging
import os

import numpy as np
import pandas as pd
import pandas._libs.json as ujson
import scipy.sparse
from cirrocumulus.anndata_dataset import AnndataDataset
from cirrocumulus.dataset_api import DatasetAPI
from cirrocumulus.io_util import get_markers, filter_markers, add_spatial, SPATIAL_HELP, unique_id
from cirrocumulus.simple_data import SimpleData

logger = logging.getLogger("cirro")

cluster_fields = ['seurat_clusters', 'leiden', 'louvain']
categorical_fields_convert = ['seurat_clusters']


def read_adata(path, backed=False, spatial_directory=None, use_raw=False):
    import anndata
    adata = anndata.read_loom(path) if path.lower().endswith('.loom') else anndata.read(path,
        backed=backed)
    if use_raw and adata.raw is not None and adata.shape[0] == adata.raw.shape[0]:
        logger.info('Using adata.raw')
        adata = anndata.AnnData(X=adata.raw.X, var=adata.raw.var, obs=adata.obs, obsm=adata.obsm, uns=adata.uns)

    if spatial_directory is not None:
        if not add_spatial(adata, spatial_directory):
            print('No spatial data found in {}'.format(spatial_directory))
    if not backed:
        if scipy.sparse.issparse(adata.X) and scipy.sparse.isspmatrix_csr(adata.X):
            adata.X = adata.X.tocsc()

    def fix_column_names(df):
        rename = {}
        for c in df.columns:
            if c.find(' ') != -1:
                rename[c] = c.replace(' ', '_')
        return df.rename(rename, axis=1) if len(rename) > 0 else df

    adata.obs = fix_column_names(adata.obs)
    adata.var = fix_column_names(adata.var)
    for field in categorical_fields_convert:
        if field in adata.obs and not pd.api.types.is_categorical_dtype(adata.obs[field]):
            logger.info('Converting {} to categorical'.format(field))
            adata.obs[field] = adata.obs[field].astype('category')
    for key in adata.obsm:
        if key.find(' ') != -1:
            new_key = key.replace(' ', '_')
            adata.obsm[new_key] = adata.obsm[key]
            del adata.obsm[key]
    return adata


def to_json(data):
    return ujson.dumps(data, double_precision=2, orient='values')


def make_unique(index, join='-1'):
    index = index.str.replace('/', '_')
    lower_index = index.str.lower()
    if lower_index.is_unique:
        return index
    from collections import defaultdict

    indices_dup = lower_index.duplicated(keep="first")
    values_dup_lc = lower_index.values[indices_dup]
    values_dup = index.values[indices_dup]
    counter = defaultdict(lambda: 0)
    for i, v in enumerate(values_dup_lc):
        counter[v] += 1
        values_dup[i] += join + str(counter[v])
    values = index.values
    values[indices_dup] = values_dup
    index = pd.Index(values)
    return index


def write_obs_stats(directory, results):
    os.makedirs(os.path.join(directory, 'stats', 'obs'), exist_ok=True)
    for column in results:
        dimension_or_measure_summary = results[column]
        if 'categories' in dimension_or_measure_summary:
            with gzip.open(os.path.join(directory, 'stats', 'obs', column + '.json.gz'), 'wt') as f:
                f.write(to_json(dimension_or_measure_summary))
        else:
            with gzip.open(os.path.join(directory, 'stats', 'obs', column + '.json.gz'), 'wt') as f:
                f.write(to_json(dimension_or_measure_summary))


def write_X_bins(base_output_dir, n_bins=25):
    # read in means from stats/X
    obs_avg = []
    names = []
    d = os.path.join(base_output_dir, 'stats', 'X')
    for file in os.listdir(d):
        if file.endswith('.json.gz'):
            with gzip.open(os.path.join(d, file), 'rt') as f:
                names.append(file[0:file.rindex('.json.gz')])
                result = json.load(f)
                obs_avg.append(result['mean'])

    obs_avg = pd.Series(np.array(obs_avg), index=names)
    n_items = int(np.round(len(obs_avg) / (n_bins - 1)))
    obs_cut = obs_avg.rank(method='min') // n_items
    obs_cut = obs_cut.astype(int)
    with gzip.open(os.path.join(base_output_dir, 'X.json.gz'), 'wt') as f:
        f.write(to_json(obs_cut.to_dict()))


def write_X_stats(directory, results):
    os.makedirs(os.path.join(directory, 'stats', 'X'), exist_ok=True)
    for column in results:
        measure_summary = results[column]
        with gzip.open(os.path.join(directory, 'stats', 'X', column + '.json.gz'), 'wt') as f:
            f.write(to_json(measure_summary))


def write_grouped_stats(directory, dotplot_results):
    # mean and fraction expressed for every gene per category
    for dotplot_result in dotplot_results:
        categories = dotplot_result['categories']
        category_name = dotplot_result['name']
        os.makedirs(os.path.join(directory, 'grouped_stats', 'obs', category_name, 'X'), exist_ok=True)
        with gzip.open(os.path.join(directory, 'grouped_stats', 'obs', category_name, 'index.json.gz'),
                'wt') as f:
            f.write(to_json(categories))
        values = dotplot_result['values']
        for value in values:
            var_field = value['name']
            with gzip.open(os.path.join(directory, 'grouped_stats', 'obs', category_name, 'X', var_field + '.json.gz'),
                    'wt') as f:
                f.write(to_json(value))


def write_basis_X(coords_group, basis_group, adata, result):
    shape = (coords_group['index'].shape[0], adata.shape[1])
    X = basis_group.require_dataset('X', shape=shape, chunks=(None, 1), dtype='f4')
    for column in result['values']:
        X_index = adata.var.index.get_indexer_for([column])[0]
        X[:, X_index] = result['values'][column]


def require_binned_basis_group(store, basis):
    basis_group = store.require_group('obsm_summary/' + basis['full_name'])
    basis_group.attrs['nbins'] = basis['nbins']
    basis_group.attrs['agg'] = basis['agg']
    basis_group.attrs['name'] = basis['name']
    basis_group.attrs['dimensions'] = basis['dimensions']
    basis_group.attrs['precomputed'] = True
    return basis_group


def write_basis_obs(basis, coords_group, obs_group, result):
    coords_group['index'] = result['bins']
    ndim = len(basis['coordinate_columns'])
    nbins = len(result['bins'])

    value_ds = coords_group.require_dataset('value', dtype='f4', shape=(nbins, ndim))

    for i in range(len(basis['coordinate_columns'])):
        coord_field = basis['coordinate_columns'][i]
        value_ds[:, i] = np.array(result['coordinates'][coord_field])
    # write_pq(bin_dict, output_directory, 'index')

    for column in result['values']:
        vals = result['values'][column]
        if not isinstance(vals, dict):  # continuous field
            obs_group[column] = vals
        else:
            g = obs_group.require_group(column)
            g['purity'] = vals['purity']
            g['mode'] = vals['value']
        # write_pq(vals, output_directory, column)


class PrepareData:

    def __init__(self, adata, output, dimensions=None, groups=[], group_nfeatures=10, markers=[],
                 output_format='parquet'):
        self.adata = adata
        self.groups = groups
        self.group_nfeatures = group_nfeatures
        self.markers = markers
        self.stats = False
        self.nbins = None
        self.bin_agg_function = None
        self.output_format = output_format

        # if basis_list is None or len(basis_list) == 0:
        #     basis_list = list(self.adata.obsm_keys())
        self.basis_list_to_precompute = None

        index = make_unique(self.adata.var.index.append(pd.Index(self.adata.obs.columns)))
        self.adata.var.index = index[0:len(self.adata.var.index)]
        self.adata.obs.columns = index[len(self.adata.var.index):]

        self.base_output_dir = output

        self.dataset_api = DatasetAPI()
        ds = AnndataDataset()
        ds.path_to_data[''] = self.adata
        self.dataset_api.add(ds)

        self.input_dataset = {'name': '', 'url': '', 'id': ''}
        dimensions_supplied = dimensions is not None and len(dimensions) > 0
        self.dimensions = [] if not dimensions_supplied else dimensions
        self.measures = []
        self.others = []
        for i in range(len(self.adata.obs.columns)):
            name = self.adata.obs.columns[i]
            c = self.adata.obs[name]
            if pd.api.types.is_object_dtype(c):
                self.adata.obs[name] = self.adata.obs[name].astype('category')
                c = self.adata.obs[name]
            if not dimensions_supplied and pd.api.types.is_categorical_dtype(c):
                if 1 < len(c.cat.categories) < 2000:
                    self.dimensions.append(name)
                    if c.isna().sum() > 0:
                        logger.info('Replacing nans in {}'.format(name))
                        self.adata.obs[name] = self.adata.obs[name].astype(str)
                        self.adata.obs.loc[self.adata.obs[name].isna(), name] = ''
                        self.adata.obs[name] = self.adata.obs[name].astype('category')
                else:
                    self.others.append(name)
            elif not pd.api.types.is_string_dtype(c) and not pd.api.types.is_object_dtype(c):
                self.measures.append('obs/' + name)
            else:
                self.others.append(name)

    def get_path(self, path):
        return os.path.join(self.base_output_dir, path)

    def execute(self):
        basis_list = self.basis_list_to_precompute
        nbins = self.nbins
        group_nfeatures = self.group_nfeatures
        bin_agg_function = self.bin_agg_function
        output_format = self.output_format
        if not os.path.exists(self.base_output_dir):
            os.makedirs(self.base_output_dir, exist_ok=True)

        schema = self.get_schema()
        results = schema.get('results', [])
        if len(results) > 0:
            uns_dir = os.path.join(self.base_output_dir, 'uns')
            os.makedirs(uns_dir, exist_ok=True)
            for i in range(len(results)):  # keep id, name, type in schema, store rest in file
                result = results[i]
                result_id = result.pop('id')
                results[i] = dict(id=result_id, name=result.pop('name'), type=result.pop('type'),
                    content_type='application/json', content_encoding='gzip')
                with gzip.open(os.path.join(uns_dir, result_id + '.json.gz'), 'wt') as f:
                    f.write(to_json(result))

        markers = schema.get('markers', [])

        if len(markers) == 0 and self.groups is None:
            groups = []
            for field in self.adata.obs.columns:
                field_lc = field.lower()
                for cluster_field in cluster_fields:
                    if field_lc.find(cluster_field) != -1:
                        groups.append(field)
                        break
            self.groups = groups
        if self.groups is not None:
            for field in self.groups:
                if not pd.api.types.is_categorical_dtype(self.adata.obs[field]):
                    self.adata.obs[field] = self.adata.obs[field].astype('category')
                if len(self.adata.obs[field].cat.categories) > 1:
                    markers += SimpleData.find_markers(self.adata, field, group_nfeatures)
            schema['markers'] = markers

        images = self.adata.uns.get('images')
        if images is not None:
            image_dir = os.path.join(self.base_output_dir, 'images')
            if not os.path.exists(image_dir):
                os.mkdir(image_dir)
            for image in images:
                path = image['image']
                import shutil
                shutil.copy(path, os.path.join(image_dir, os.path.basename(path)))
                image['image'] = 'images/' + os.path.basename(path)

        if output_format == 'parquet':
            from cirrocumulus.parquet_io import save_adata_pq
            save_adata_pq(self.adata, schema, self.base_output_dir)
        elif output_format == 'json':
            from cirrocumulus.json_io import save_adata_json
            save_adata_json(self.adata, schema, self.base_output_dir)
        elif output_format == 'jsonl':
            from cirrocumulus.jsonl_io import save_adata_jsonl
            save_adata_jsonl(self.adata, schema, self.base_output_dir)
        else:
            raise ValueError("Unknown format")

        # if self.stats:
        #     self.summary_stats()
        #     self.grouped_stats()

        if nbins is not None and basis_list is not None:
            for basis_name in basis_list:
                self.grid_embedding(basis_name, bin_agg_function, nbins)


    # def summary_stats(self):
    #     dimensions = self.dimensions
    #     measures = self.measures
    #     dataset_api = self.dataset_api
    #     input_dataset = self.input_dataset
    #     X_range = self.X_range
    #     adata = self.adata
    #     base_dir = self.base_output_dir
    #
    #     if X_range[0] == 0:
    #         process_results = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset,
    #             stats=dict(measures=measures, dimensions=dimensions))
    #         write_obs_stats(base_dir, process_results['summary'])
    #
    #     logger.info('Summary stats X {}-{}'.format(X_range[0], X_range[1]))
    #     process_results = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset,
    #         stats=dict(measures=adata.var_names[X_range[0]:X_range[1]], dimensions=[]))
    #     write_X_stats(base_dir, process_results['summary'])
    #     write_X_bins(base_dir)
    #
    # # stats, grouped by categories in adata.obs for dot plot
    # def grouped_stats(self):
    #     dimensions = self.dimensions
    #     dataset_api = self.dataset_api
    #     input_dataset = self.input_dataset
    #     adata = self.adata
    #     X_range = self.X_range
    #     base_dir = self.base_output_dir
    #     logger.info('Grouped stats X {}-{}'.format(X_range[0], X_range[1]))
    #     process_results = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset,
    #         grouped_stats=dict(measures=adata.var_names[X_range[0]:X_range[1]], dimensions=dimensions))
    #     dotplot_results = process_results['dotplot']
    #
    #     write_grouped_stats(base_dir, dotplot_results)


    # def grid_embedding(self, basis_name, summary, nbins):
    #     if nbins <= 0:
    #         return
    #     logger.info('{} embedding'.format(basis_name))
    #     basis = get_basis(basis_name, nbins, summary, 2, False)
    #
    #     dimensions = self.dimensions
    #     measures = self.measures
    #     dataset_api = self.dataset_api
    #     input_dataset = self.input_dataset
    #     adata = self.adata
    #     X_range = self.X_range
    #     full_basis_name = basis['full_name']
    #
    #     basis_group = require_binned_basis_group(store, basis)
    #     obs_group = basis_group.require_group('obs')
    #     coords_group = basis_group.require_group('coords')
    #
    #     if X_range[0] == 0:
    #         df_with_coords = pd.DataFrame()
    #         obsm = adata.obsm[basis_name]
    #         for i in range(len(basis['coordinate_columns'])):
    #             df_with_coords[basis['coordinate_columns'][i]] = obsm[:, i]
    #         EmbeddingAggregator.convert_coords_to_bin(df_with_coords, nbins, basis['coordinate_columns'],
    #             bin_name=full_basis_name)
    #         coords_group['bins'] = df_with_coords[full_basis_name].values
    #         coords_group['bin_coords'] = df_with_coords[basis['coordinate_columns']].values
    #
    #         result = data_processing.handle_data(dataset_api=dataset_api, dataset=input_dataset, basis=basis,
    #             measures=measures + ['__count'], dimensions=dimensions, quick=False)
    #         write_basis_obs(basis, coords_group, obs_group, result)
    #         logger.info('{} embedding finished writing obs'.format(basis_name))
    #     # write X to obsm_summary/name
    #     logger.info('{} embedding X {}-{}'.format(basis_name, X_range[0], X_range[1]))
    #     result = data_processing.handle_embedding(dataset_api=dataset_api, dataset=input_dataset, basis=basis,
    #         measures=adata.var_names[X_range[0]:X_range[1]], dimensions=[])
    #     write_basis_X(coords_group, basis_group, adata, result)
    #     # write_pq(dict(value=result['values'][column]), output_directory, column)

    def get_schema(self):
        basis_list = self.basis_list_to_precompute
        nbins = self.nbins
        bin_agg_function = self.bin_agg_function
        result = SimpleData.schema(self.adata)
        markers = result.get('markers', [])

        if self.markers is not None:  # add from file
            markers += get_markers(self.markers)

        markers = filter_markers(self.adata, markers)

        for marker in markers:
            if marker.get('id') is None:
                marker['id'] = unique_id()
            marker['readonly'] = True
        result['markers'] = markers
        result['format'] = self.output_format
        if self.stats:
            result['precomputed'] = True  # has stats
        if nbins is not None:
            result['embeddings'] = []
            for basis_name in basis_list:
                result['embeddings'].append(
                    {'precomputed': True, 'name': basis_name, 'nbins': nbins, 'agg': bin_agg_function,
                     'dimensions': 2})
        return result


def main(argsv):
    parser = argparse.ArgumentParser(
        description='Prepare a dataset for cirrocumulus server.')
    parser.add_argument('dataset', help='Path to a h5ad, loom, or Seurat file')
    parser.add_argument('--out', help='Path to output directory')
    # parser.add_argument('--stats', dest="stats", help='Generate precomputed stats', action='store_true')
    parser.add_argument('--backed', help='Load h5ad file in backed mode', action='store_true')
    # parser.add_argument('--basis', help='List of embeddings to precompute', action='append')
    #  parser.add_argument('--groups', help='List of groups to precompute summary statistics', action='append')
    parser.add_argument('--markers',
        help='Path to JSON file of precomputed markers that maps name to features. For example {"a":["gene1", "gene2"], "b":["gene3"]',
        action='append')
    parser.add_argument('--groups',
        help='List of groups to compute markers for (e.g. louvain). Note that markers created with scanpy or cumulus are automatically included.',
        action='append')
    parser.add_argument('--group_nfeatures', help='Number of marker genes/features to include', type=int, default=10)
    parser.add_argument('--spatial', help=SPATIAL_HELP)
    # parser.add_argument('--output_format', help='Output file format', choices=['json', 'parquet'], default='parquet')
    # parser.add_argument('--nbins', help='Number of bins. Set to 0 to disable binning', default=500, type=int)
    # parser.add_argument('--summary', help='Bin summary statistic for numeric values', default='max')
    # parser.add_argument('--X_range', help='Start and end position of data matrix (e.g. 0-5000)', type=str)

    args = parser.parse_args(argsv)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())

    input_basis = []  # check if specified as comma delimited list
    input_X_range = None
    out = args.out
    input_dataset = args.dataset
    output_format = 'parquet'  # args.output_format
    if out is None:
        out = os.path.basename(input_dataset)
        out = out[0:out.rindex('.')]
    tmp_file = None
    use_raw = False
    if input_dataset.lower().endswith('.rds'):
        import subprocess
        import tempfile
        import pkg_resources

        _, h5_file = tempfile.mkstemp(suffix='.h5ad')
        os.remove(h5_file)
        subprocess.check_call(
            ['Rscript', pkg_resources.resource_filename("cirrocumulus", 'seurat2h5ad.R'), input_dataset, h5_file])
        input_dataset = h5_file
        tmp_file = h5_file
        use_raw = True
    # summary = 'max'
    # nbins = -1
    # summary = args.summary
    # if args.basis is not None:
    #     for b in args.basis:
    #         input_basis.extend(b.split(','))
    # input_X_range = args.X_range
    # if input_X_range is not None:
    #     input_X_range = input_X_range.split('-')
    #     assert len(input_X_range) == 2
    #     input_X_range[0] = int(input_X_range[0])
    #     input_X_range[1] = int(input_X_range[1])

    adata = read_adata(input_dataset, backed=args.backed, spatial_directory=args.spatial, use_raw=use_raw)
    prepare_data = PrepareData(adata=adata, output=out, dimensions=args.groups, groups=args.groups,
        group_nfeatures=args.group_nfeatures,
        markers=args.markers, output_format=output_format)
    prepare_data.execute()
    if tmp_file is not None:
        os.remove(tmp_file)


if __name__ == '__main__':
    import sys

    main(sys.argv)



import os

from cirrocumulus.anndata_dataset import AnndataDataset
from cirrocumulus.io_util import get_markers, filter_markers, add_spatial, SPATIAL_HELP
from cirrocumulus.parquet_dataset import ParquetDataset


def configure_app(app, list_of_dataset_paths, spatial_directories, backed, marker_paths):
    from cirrocumulus.api import dataset_api
    from cirrocumulus.local_db_api import LocalDbAPI
    from cirrocumulus.no_auth import NoAuth

    app.config['AUTH'] = NoAuth()
    dataset_api.add(ParquetDataset())
    anndata_dataset = AnndataDataset('r' if backed else None)
    dataset_ids = []
    for dataset_paths in list_of_dataset_paths:
        dataset_paths = dataset_paths.split(',')
        dataset_id = os.path.normpath(dataset_paths[0])
        dataset_ids.append(dataset_id)
        if len(dataset_paths) > 1:
            to_concat = []
            all_ids = None
            for path in dataset_paths:
                d = anndata_dataset.get_data(path)
                all_ids = d.obs.index.union(all_ids) if all_ids is not None else d.obs.index
                to_concat.append(d)
            for i in range(len(to_concat)):
                d = to_concat[i]
                missing_ids = all_ids.difference(d.obs.index)
                if len(missing_ids) > 0:
                    import scipy.sparse
                    import anndata
                    import pandas as pd
                    X = None
                    if d.shape[1] > 0:
                        empty = scipy.sparse.csr_matrix((len(missing_ids), d.shape[1]))
                        X = scipy.sparse.vstack((d.X, empty), format='csr')
                    missing_df = pd.DataFrame(index=missing_ids)
                    for column in d.obs:
                        if pd.api.types.is_bool_dtype(d.obs[column]):
                            missing_df[column] = False

                    obs = pd.concat((d.obs, missing_df))
                    # for column in d.obs:
                    #     if pd.api.types.is_categorical_dtype(d.obs[column]):
                    #         obs[column] = obs[column].astype('category')
                    d = anndata.AnnData(X=X, obs=obs, var=d.var)
                d = d[all_ids]  # same order
                to_concat[i] = d
            X_list = []
            obs = None
            obsm = {}
            var = None
            for d in to_concat:
                if d.shape[1] > 0:
                    X_list.append(d.X)
                    var = pd.concat((var, d.var)) if var is not None else d.var
                obs = obs.join(d.obs) if obs is not None else d.obs
                for key in d.obsm_keys():
                    obsm[key] = d.obsm[key]

            X = scipy.sparse.hstack(X, format='csr') if len(X_list) > 1 else X_list[0]
            adata = anndata.AnnData(X=X, obs=obs, var=var, obsm=obsm)
            adata.var_names_make_unique()
            anndata_dataset.add_data(dataset_id, adata)
        dataset_api.add(anndata_dataset)

    app.config['DATABASE'] = LocalDbAPI(dataset_ids)

    if spatial_directories is not None and len(spatial_directories) > 0:
        for i in range(len(spatial_directories)):
            spatial_directory = spatial_directories[i]
            if spatial_directory != '':
                adata = anndata_dataset.get_data(dataset_ids[i])
                if not add_spatial(adata, spatial_directory):
                    print('No spatial data found in {}'.format(spatial_directory))

    if marker_paths is not None and len(marker_paths) > 0:
        markers = get_markers(marker_paths)
        for dataset_id in dataset_ids:
            d = anndata_dataset.get_data(dataset_id)
            existing_markers = d.uns.get('markers', [])
            markers += existing_markers
            # remove genes in dict that are not in dataset
            d.uns['markers'] = filter_markers(d, markers)


def create_app():
    from cirrocumulus.api import blueprint
    from flask_compress import Compress
    from flask import Flask, send_from_directory
    os.environ['WERKZEUG_RUN_MAIN'] = 'true'
    app = Flask(__name__, static_folder='client', static_url_path='')
    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0
    app.register_blueprint(blueprint, url_prefix='/api')

    @app.route('/')
    def root():
        return send_from_directory(os.path.abspath(os.path.join(app.root_path, "client")), "index.html")

    Compress(app)
    return app


def main(argsv):
    import argparse
    parser = argparse.ArgumentParser(description='Run cirrocumulus')
    parser.add_argument('dataset',
        help='Path to dataset in h5ad, loom, or STAR-Fusion format. Separate multiple datasets with '
             'a comma instead of a space in order to join datasets by cell id', nargs='+')
    parser.add_argument('--spatial', help=SPATIAL_HELP, nargs='*')
    parser.add_argument('--markers',
        help='Path to JSON file that maps name to features. For example {"a":["gene1", "gene2"], "b":["gene3"]}',
        nargs='*')
    parser.add_argument('--backed', help='Load h5ad file in backed mode', action='store_true')
    parser.add_argument('--host',
        help='Host IP address')  # set to 0.0.0.0 to make it accessible from other computers WITHOUT SECURITY.

    parser.add_argument('--port', help='Server port', default=5000, type=int)
    parser.add_argument('--no-open', dest='no_open', help='Do not open your web browser', action='store_true')

    args = parser.parse_args(argsv)
    app = create_app()
    # from flask_cors import CORS
    # CORS(app)
    configure_app(app, args.dataset, args.spatial, args.backed, args.markers)
    if not args.no_open:
        import webbrowser
        host = args.host if args.host is not None else 'http://127.0.0.1'
        url = host + ':' + str(args.port)
        webbrowser.open(url)
    app.run(host=args.host, port=args.port, debug=False)


if __name__ == "__main__":
    main()



import gzip
import json
import os
from abc import abstractmethod, ABC


class AbstractDataset(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_suffixes(self):
        pass

    @abstractmethod
    def read_dataset(self, file_system, path, keys=None, dataset=None):
        pass


    def has_precomputed_stats(self, file_system, path, dataset):
        return file_system.exists(os.path.join(path, 'stats'))

    def read_precomputed_stats(self, file_system, path, obs_keys=[], var_keys=[]):
        result = {}
        if len(obs_keys) > 0:
            for key in obs_keys:
                with gzip.open(file_system.open(path + '/stats/obs/' + key + '.json.gz', 'rb')) as f:
                    result[key] = json.load(f)

        if len(var_keys) > 0:
            for key in var_keys:
                with gzip.open(file_system.open(path + '/stats/X/' + key + '.json.gz', 'rb')) as f:
                    result[key] = json.load(f)
        return result

    def read_precomputed_grouped_stats(self, file_system, path, obs_keys=[], var_keys=[]):
        results = []
        if len(obs_keys) > 0 and len(var_keys) > 0:

            for obs_key in obs_keys:
                values = []
                with gzip.open(file_system.open(path + '/grouped_stats/obs/' + obs_key + '/index.json.gz', 'rb')) as f:
                    categories = json.load(f)
                result = {'categories': categories, 'name': obs_key, 'values': values}
                results.append(result)
                for var_key in var_keys:
                    with gzip.open(
                            file_system.open(path + '/grouped_stats/obs/' + obs_key + '/X/' + var_key + '.json.gz',
                                'rb')) as f:
                        r = json.load(f)
                        r['name'] = var_key
                        values.append(r)
        return results

    def schema(self, file_system, path):
        if path.endswith('.gz'):
            import gzip
            with gzip.open(file_system.open(path)) as s:
                return json.load(s)
        else:
            with file_system.open(path) as s:
                return json.load(s)



import os

from cirrocumulus.envir import CIRRO_AUTH_CLIENT_ID, CIRRO_DB_URI, CIRRO_DATABASE, CIRRO_EMAIL, CIRRO_SERVE, \
    CIRRO_FOOTER, CIRRO_UPLOAD, CIRRO_BRAND
from cirrocumulus.launch import create_app

app = None

DEFAULT_DB_URI = 'mongodb://localhost:27017/'
DEFAULT_DATABASE = 'cirrocumulus'


def cached_app():
    global app
    if app is None:
        app = create_app()
        # from flask_cors import CORS
        # CORS(app)
        configure_app(app)
    return app


def configure_app(app):
    from cirrocumulus.api import dataset_api
    from cirrocumulus.no_auth import NoAuth
    auth_client_id = os.environ.get(CIRRO_AUTH_CLIENT_ID)
    db_uri = os.environ.get(CIRRO_DB_URI, DEFAULT_DB_URI)
    database = os.environ.get(CIRRO_DATABASE, DEFAULT_DATABASE)
    email = os.environ.get(CIRRO_EMAIL)
    os.environ[CIRRO_SERVE] = 'true'
    if auth_client_id is None:
        app.config['AUTH'] = NoAuth()
    else:
        from cirrocumulus.google_auth import GoogleAuth
        app.config['AUTH'] = GoogleAuth(auth_client_id)
    from cirrocumulus.mongo_db import MongoDb
    app.config['DATABASE'] = MongoDb(db_uri, database, email)
    from cirrocumulus.parquet_dataset import ParquetDataset
    dataset_api.add(ParquetDataset())
    from cirrocumulus.anndata_dataset import AnndataDataset
    anndata_dataset = AnndataDataset('r' if False else None)
    dataset_api.add(anndata_dataset)


def main(argsv):
    import argparse
    import os
    parser = argparse.ArgumentParser(description='Run cirrocumulus server')
    parser.add_argument('--database', help='MongoDB database', default=DEFAULT_DATABASE)
    parser.add_argument('--db_uri', help='MongoDB database connection URI', default=DEFAULT_DB_URI)
    parser.add_argument('--email', help='Email address that server runs as')
    parser.add_argument('--auth_client_id', help='OAuth client id')
    parser.add_argument('-w', '--workers', dest='workers', help='The number of worker processes', type=int)
    parser.add_argument('-t', '--timeout', dest='timeout',
        help='Workers silent for more than this many seconds are killed and restarted', type=int, default=30)
    parser.add_argument('-b', '--bind', dest='bind',
        help='Server socket to bind. Server sockets can be any of $(HOST), $(HOST):$(PORT), fd://$(FD), or unix:$(PATH). An IP is a valid $(HOST).',
        default='127.0.0.1:5000')
    parser.add_argument('--footer', help='Markdown file to customize the application footer')
    parser.add_argument('--header', help='Markdown file to customize the application header')
    parser.add_argument('--upload', help='URL to allow users to upload files')

    args = parser.parse_args(argsv)

    bind = args.bind if args.bind is not None else '127.0.0.1:5000'
    if args.auth_client_id is not None:
        os.environ[CIRRO_AUTH_CLIENT_ID] = args.auth_client_id

    if args.footer is not None:
        os.environ[CIRRO_FOOTER] = args.footer

    if args.header is not None:
        os.environ[CIRRO_BRAND] = args.header

    if args.email is not None:
        os.environ[CIRRO_EMAIL] = args.email
    if args.db_uri is not None:
        os.environ[CIRRO_DB_URI] = args.db_uri
    if args.database is not None:
        os.environ[CIRRO_DATABASE] = args.database
    if args.workers is not None:
        workers = args.workers
    else:
        import os
        workers = 2 * os.cpu_count()
    if args.upload is not None:
        os.environ[CIRRO_UPLOAD] = args.upload

    run_args = [
            'gunicorn',
            '-b', bind,
            '-w', str(workers),
            '-t', str(args.timeout),
            '-n', 'cirrocumulus-webserver',
            'cirrocumulus.serve:cached_app()'
    ]
    import subprocess
    subprocess.check_call(run_args)


if __name__ == "__main__":
    main()



import pandas as pd
import scipy.sparse

from cirrocumulus.dotplot_aggregator import DotPlotAggregator
from cirrocumulus.embedding_aggregator import EmbeddingAggregator, get_basis
from cirrocumulus.feature_aggregator import FeatureAggregator
from cirrocumulus.ids_aggregator import IdsAggregator
from cirrocumulus.unique_aggregator import UniqueAggregator


def apply_filter(df, data_filter):
    keep_expr = get_filter_expr(df, data_filter)
    return df[keep_expr] if keep_expr is not None else df


def get_filter_expr(df, data_filter):
    if df is None:
        raise ValueError('df is None')

    keep_expr = None
    if data_filter is not None:
        user_filters = data_filter.get('filters', [])
        combine_filters = data_filter.get('combine', 'and')
        for filter_obj in user_filters:
            field = filter_obj[0]
            op = filter_obj[1]
            value = filter_obj[2]
            if isinstance(field, dict):  # selection box

                selected_points_basis = get_basis(field['basis'], field.get('nbins'),
                    field.get('agg'), field.get('ndim', '2'), field.get('precomputed', False))

                if 'points' in value:
                    p = value['points']
                    field = selected_points_basis['full_name'] if selected_points_basis[
                                                                      'nbins'] is not None else 'index'
                    if field == 'index':
                        keep = df.index.isin(p)
                    else:
                        keep = df[field].isin(p)
                else:
                    keep = None
                    for p in value['path']:
                        if 'z' in p:  # 3d
                            selection_keep = \
                                (df[selected_points_basis['coordinate_columns'][0]] >= p['x']) & \
                                (df[selected_points_basis['coordinate_columns'][0]] <= p['x'] + p['width']) & \
                                (df[selected_points_basis['coordinate_columns'][1]] >= p['y']) & \
                                (df[selected_points_basis['coordinate_columns'][1]] <= p['y'] + p[
                                    'height']) & \
                                (df[selected_points_basis['coordinate_columns'][2]] >= p['z']) & \
                                (df[selected_points_basis['coordinate_columns'][2]] <= p['z'] + p['depth'])
                        else:
                            selection_keep = \
                                (df[selected_points_basis['coordinate_columns'][0]] >= p['x']) & \
                                (df[selected_points_basis['coordinate_columns'][0]] <= p['x'] + p['width']) & \
                                (df[selected_points_basis['coordinate_columns'][1]] >= p['y']) & \
                                (df[selected_points_basis['coordinate_columns'][1]] <= p['y'] + p['height'])

                    keep = selection_keep | keep if keep is not None else selection_keep

            elif field == '__index':
                import numpy as np
                keep = np.zeros(len(df), dtype=bool)
                keep[value] = True
            else:
                series = df[field]
                if op == 'in':
                    keep = (series.isin(value)).values
                elif op == '>':
                    keep = (series > value)
                elif op == '=':
                    keep = (series == value)
                elif op == '<':
                    keep = (series < value)
                elif op == '!=':
                    keep = (series != value)
                elif op == '>=':
                    keep = (series >= value)
                elif op == '<=':
                    keep = (series <= value)
                else:
                    raise ValueError('Unknown filter')

            if scipy.sparse.issparse(keep):
                keep = keep.toarray().flatten()
            if hasattr(keep, 'sparse'):
                keep = keep.sparse.to_dense()
            if isinstance(keep, pd.Series):
                keep = keep.values
            if keep_expr is not None:
                if combine_filters == 'and':
                    keep_expr = keep_expr & keep
                else:
                    keep_expr = keep_expr | keep
            else:
                keep_expr = keep

    return keep_expr


def precomputed_summary(dataset_api, dataset, obs_measures, var_measures, dimensions):
    if '__count' in var_measures:
        var_measures.remove('__count')
    return dataset_api.read_precomputed_stats(dataset, obs_keys=dimensions + obs_measures, var_keys=var_measures)


def precomputed_grouped_stats(dataset_api, dataset, var_measures, dimensions):
    if (len(var_measures)) > 0 and len(dimensions) > 0:
        return dataset_api.read_precomputed_grouped_stats(dataset,
            var_keys=var_measures, obs_keys=dimensions)
    return []


def precomputed_embedding(dataset_api, dataset, basis, obs_measures, var_measures,
                          dimensions):
    if (len(obs_measures) + len(var_measures) + len(dimensions)) == 0:
        obs_measures = ['__count']
    return dataset_api.read_precomputed_basis(dataset, obs_keys=obs_measures + dimensions, var_keys=var_measures,
        basis=basis)


def get_var_name_type(key):
    index = key.find('/')
    if index == -1:
        return key, 'X'
    else:
        key_type = key[0:index]
        name = key[index + 1:]
        return name, key_type


def get_type_to_measures(measures):
    type2measures = dict(X=[], obs=[])
    for measure in measures:
        name, key_type = get_var_name_type(measure)
        type_measures = type2measures.get(key_type)
        if type_measures is None:
            type_measures = []
            type2measures[key_type] = type_measures
        type_measures.append(name)
    return type2measures


def check_bin_input(nbins):
    if nbins is not None:
        nbins = int(nbins)
        nbins = min(1000, nbins)
        if nbins <= 0:
            nbins = None
    return nbins


def handle_export_dataset_filters(dataset_api, dataset, data_filters):
    import json

    reformatted_filters = []
    filter_names = []
    for data_filter_obj in data_filters:
        filter_value = json.loads(data_filter_obj['value'])
        filter_names.append(data_filter_obj['name'])
        reformatted_filters.append(filter_value)

    df = get_df(dataset_api, dataset, measures=['obs/index'], data_filters=reformatted_filters)
    result_df = pd.DataFrame(index=df['index'])
    for i in range(len(reformatted_filters)):
        data_filter = reformatted_filters[i]
        filter_name = filter_names[i]
        df_filtered = apply_filter(df, data_filter)
        df = pd.DataFrame(index=df_filtered['index'])
        df[filter_name] = True
        result_df = result_df.join(df, rsuffix='r')
    result_df.fillna(False, inplace=True)
    return result_df.to_csv()


# embedding - list of basis and coords (whether to return coordinates). For binned embeddings, also measures and dimensions.
def handle_data(dataset_api, dataset, embedding_list=None, values=None, grouped_stats=None, stats=None, selection=None):
    dimensions = set()
    measures = set()
    basis_list = []
    basis_keys = set()
    if embedding_list is not None:
        for embedding in embedding_list:
            precomputed = embedding.get('precomputed', False)
            nbins = check_bin_input(embedding.get('nbins', None))
            agg_function = embedding.get('agg', 'max')
            ndim = embedding.get('ndim', '2')
            coords = embedding.get('coords', True)
            basis = get_basis(embedding.get('basis'), nbins=nbins, agg=agg_function, dimensions=ndim,
                precomputed=precomputed, coords=coords)
            if nbins is not None:
                dimensions.update(embedding.get('dimensions', []))
                measures.update(embedding.get('measures', []))

            if basis['full_name'] not in basis_keys:
                basis_list.append(basis)
                basis_keys.add(basis['full_name'])

            embedding['basis'] = basis
    if values is not None:
        dimensions.update(values.get('dimensions', []))
        measures.update(values.get('measures', []))

    if selection is not None:
        data_filter = selection.get('filter')
        var_keys_filter, obs_keys_filter, selected_points_filter_basis_list = data_filter_keys(data_filter)
        selection['basis'] = selected_points_filter_basis_list
        measures.update(var_keys_filter)
        dimensions.update(obs_keys_filter)
        dimensions.update(selection.get('dimensions', []))
        measures.update(selection.get('measures', []))
        selection_embeddings = selection.get('embeddings', [])
        embeddings_mapped = []
        for embedding in selection_embeddings:
            basis = get_basis(embedding.get('basis'), nbins=embedding.get('nbins'), agg=embedding.get('agg', 'max'),
                dimensions=embedding.get('ndim', '2'),
                precomputed=embedding.get('precomputed', False))
            embeddings_mapped.append(basis)
        selection['embeddings'] = embeddings_mapped
        for basis_obj in selected_points_filter_basis_list + embeddings_mapped:
            if basis_obj['full_name'] not in basis_keys:
                basis_list.append(basis_obj)
                basis_keys.add(basis_obj['full_name'])

    if grouped_stats is not None:
        grouped_stats_dimensions = grouped_stats.get('dimensions', [])
        for d in grouped_stats_dimensions:
            if isinstance(d, list):
                dimensions.update(d)
            else:
                dimensions.add(d)
        measures.update(grouped_stats.get('measures', []))
    if stats is not None:
        dimensions.update(stats.get('dimensions', []))
        measures.update(stats.get('measures', []))

    keys = get_type_to_measures(measures)
    keys['obs'] += list(dimensions)
    keys['basis'] = basis_list
    df = dataset_api.read_dataset(dataset=dataset, keys=keys)
    for basis_obj in basis_list:
        if not basis_obj['precomputed'] and basis_obj['nbins'] is not None:
            EmbeddingAggregator.convert_coords_to_bin(df=df,
                nbins=basis_obj['nbins'],
                bin_name=basis_obj['full_name'],
                coordinate_columns=basis_obj['coordinate_columns'])
    results = {}
    if values is not None:
        dimensions = values.get('dimensions', [])
        measures = values.get('measures', [])
        type2measures = get_type_to_measures(measures)
        measures = []
        for v in type2measures.values():
            measures += v

        result = EmbeddingAggregator(
            measures=measures,
            dimensions=dimensions,
            nbins=None,
            basis=None,
            coords=False,
            agg_function=None,
            quick=True).execute(df)
        results['values'] = result['values']

    if embedding_list is not None:
        results['embeddings'] = []
        for embedding in embedding_list:
            dimensions = embedding.get('dimensions', [])
            measures = embedding.get('measures', [])
            type2measures = get_type_to_measures(measures)
            basis = embedding['basis']
            if basis['precomputed']:
                result = precomputed_embedding(dataset_api, dataset, basis, type2measures['obs'],
                    type2measures['X'], dimensions)
            else:
                result = EmbeddingAggregator(
                    measures=type2measures['obs'] + type2measures['X'],
                    dimensions=dimensions,
                    nbins=basis['nbins'],
                    basis=basis,
                    coords=basis['coords'],
                    agg_function=basis['agg'],
                    quick=True).execute(df)
            results['embeddings'].append(result)
    if grouped_stats is not None:
        if dataset_api.has_precomputed_stats(dataset):
            results['distribution'] = precomputed_grouped_stats(dataset_api, dataset, grouped_stats.get('measures', []),
                grouped_stats.get('dimensions', []))
        else:
            results['distribution'] = DotPlotAggregator(var_measures=grouped_stats.get('measures', []),
                dimensions=grouped_stats.get('dimensions', [])).execute(df)
    if stats is not None:
        dimensions = stats.get('dimensions', [])
        measures = stats.get('measures', [])
        type2measures = get_type_to_measures(measures)
        if dataset_api.has_precomputed_stats(dataset):
            result['summary'] = precomputed_summary(dataset_api, dataset, type2measures['obs'],
                type2measures['X'], dimensions)
        else:
            results['summary'] = FeatureAggregator(type2measures['obs'], type2measures['X'],
                dimensions).execute(df)
    if selection is not None:
        results['selection'] = {}
        dimensions = selection.get('dimensions', [])
        measures = selection.get('measures', [])
        type2measures = get_type_to_measures(measures)
        # basis_list = selection.get('basis', [])
        selection_embeddings = selection.get('embeddings', [])
        df = apply_filter(df, data_filter)
        if len(selection_embeddings) > 0:
            results['selection']['coordinates'] = {}
            for embedding in selection_embeddings:
                results['selection']['coordinates'][embedding['full_name']] = UniqueAggregator(
                    embedding['full_name'] if embedding['nbins'] is not None else 'index').execute(df)
        var_measures = type2measures['X']
        results['selection']['summary'] = FeatureAggregator(type2measures['obs'], type2measures['X'],
            dimensions).execute(df)
        if len(dimensions) > 0 and len(var_measures) > 0:
            results['selection']['distribution'] = DotPlotAggregator(var_measures=var_measures,
                dimensions=[dimensions]).execute(df)
        results['selection']['count'] = df.shape[0]

    return results


def handle_selection_ids(dataset_api, dataset, data_filter):
    df = get_selected_data(dataset_api, dataset, measures=['obs/index'], data_filter=data_filter)
    return {'ids': IdsAggregator().execute(df)}


def get_df(dataset_api, dataset, embeddings=[], measures=[], dimensions=[], data_filters=[]):
    type2measures = get_type_to_measures(measures)
    var_keys_filter = []
    obs_keys_filter = []
    selected_points_filter_basis_list = []
    for data_filter in data_filters:
        _var_keys_filter, _obs_keys_filter, selected_points_filter_basis_list = data_filter_keys(data_filter)
        selected_points_filter_basis_list += selected_points_filter_basis_list
        var_keys_filter += _var_keys_filter
        obs_keys_filter += _obs_keys_filter
    obs_keys = list(set(type2measures['obs'] + obs_keys_filter + dimensions))
    var_keys = list(set(type2measures['X'] + var_keys_filter))
    basis_objs = []
    basis_keys = set()
    for embedding in embeddings:
        basis_obj = get_basis(embedding['basis'], embedding.get('nbins'), embedding.get('agg'),
            embedding.get('ndim', '2'), embedding.get('precomputed', False))
        if basis_obj['full_name'] not in basis_keys:
            basis_objs.append(basis_obj)
            basis_keys.add(basis_obj['full_name'])
    for basis_obj in selected_points_filter_basis_list:
        if basis_obj['full_name'] not in basis_keys:
            basis_objs.append(basis_obj)
            basis_keys.add(basis_obj['full_name'])
    df = dataset_api.read_dataset(dataset=dataset, keys=dict(obs=obs_keys, X=var_keys, basis=basis_objs))
    for basis_obj in basis_objs:
        if not basis_obj['precomputed'] and basis_obj['nbins'] is not None:
            EmbeddingAggregator.convert_coords_to_bin(df=df,
                nbins=basis_obj['nbins'],
                bin_name=basis_obj['full_name'],
                coordinate_columns=basis_obj['coordinate_columns'])
    return df


def get_selected_data(dataset_api, dataset, embeddings=[], measures=[], dimensions=[], data_filter=None):
    df = get_df(dataset_api, dataset, embeddings, measures, dimensions,
        [data_filter] if data_filter is not None else [])
    df = apply_filter(df, data_filter)
    return df


def data_filter_keys(data_filter):
    var_keys = set()
    obs_keys = set()
    basis_list = []
    if data_filter is not None:
        user_filters = data_filter.get('filters', [])

        for i in range(len(user_filters)):
            user_filter = user_filters[i]
            key = user_filter[0]
            if isinstance(key, dict):
                basis = get_basis(key['basis'], key.get('nbins'), key.get('agg'),
                    key.get('ndim', '2'), key.get('precomputed', False))
                basis_list.append(basis)
            elif key == '__index':
                continue
            else:
                name, key_type = get_var_name_type(key)
                user_filter[0] = name
                if key_type == 'X':
                    var_keys.add(name)
                else:
                    obs_keys.add(name)
    return list(var_keys), list(obs_keys), basis_list



import numpy as np
import pandas as pd
import scipy.sparse

from cirrocumulus.data_processing import handle_data


def diff_results(summarized_df, measures, results):
    dotplot_result = results[0]
    values = dotplot_result['values']
    for key in measures:
        index = -1
        for i in range(len(values)):
            if values[i]['name'] == key:
                index = i
                break
        if index == -1:
            raise ValueError(key + ' not found')
        np.testing.assert_allclose(summarized_df[key]['mean'].values, values[index]['mean'],
            atol=0.000001, err_msg='{} mean'.format(key))
        np.testing.assert_allclose(summarized_df[key]['percentExpressed'].values,
            values[index]['percentExpressed'], err_msg='{} percentExpressed'.format(key))


def get_df(test_data, measures, by):
    X = test_data[:, measures].X
    if scipy.sparse.issparse(X):
        X = X.toarray()
    df = pd.DataFrame(data=X, columns=measures)
    df[by] = test_data.obs[by].values

    def percentExpressed(g):
        return 100 * ((g != 0).sum() / len(g))

    return df.groupby(by, observed=True).agg(['mean', percentExpressed])


def test_dot_plot(dataset_api, input_dataset, test_data, measures, by):
    process_results = handle_data(dataset_api=dataset_api, dataset=input_dataset, grouped_stats=dict(measures=measures,
        dimensions=[by]))
    dotplot_result = process_results['distribution']
    diff_results(get_df(test_data, measures, by), measures, dotplot_result)


# def test_save_grouped_stats(tmp_path, dataset_api, test_data, measures, by):
#     import zarr
#     url = str(tmp_path) + '.zarr'
#     make_ordered(test_data.obs, None)
#     test_data = test_data[:, measures]
#     test_data.write_zarr(url)
#     test_data[:, measures].write_zarr(url)
#     store = zarr.open(url)
#     results = DotPlotAggregator(var_measures=measures, dimensions=[by]).execute(test_data)
#     write_grouped_stats(store, test_data, results)
#     input_dataset = Entity(url, {'name': tmp_path, 'url': url})
#     result = dataset_api.read_precomputed_grouped_stats(input_dataset, obs_keys=[by], var_keys=measures)
#     diff_results(get_df(test_data, measures, by), measures, result)



import datetime
import json
import os

from cirrocumulus.io_util import unique_id


def create_dataset_meta(path):
    result = {'id': path, 'url': path, 'name': os.path.splitext(os.path.basename(path))[0], 'description': ''}
    if os.path.basename(path).endswith('.json'):
        with open(path, 'rt') as f:
            result.update(json.load(f))

    return result


def write_json(json_data, json_path):
    with open(json_path, 'wt') as f:
        json.dump(json_data, f)


class LocalDbAPI:

    def __init__(self, paths):
        self.dataset_to_info = {}  # json_data, meta, json_path
        self.job_id_to_job = {}

        for path in paths:
            json_data = {}
            basename = os.path.splitext(path)[0]
            old_path = basename + '_filters.json'
            json_path = basename + '.json'
            if os.path.exists(old_path) and os.path.getsize(old_path) > 0:
                with open(old_path, 'rt') as f:
                    json_data['filters'] = json.load(f)

            if os.path.exists(json_path) and os.path.getsize(json_path) > 0:
                with open(json_path, 'rt') as f:
                    json_data.update(json.load(f))
            meta = create_dataset_meta(path)
            if 'filters' not in json_data:
                json_data['filters'] = {}
            if 'categories' not in json_data:
                json_data['categories'] = {}
            self.dataset_to_info[path] = dict(json_data=json_data, meta=meta, json_path=json_path)

    def server(self):
        return dict(mode='client')

    def user(self, email):
        return dict()

    def datasets(self, email):
        results = []
        for key in self.dataset_to_info:
            results.append(self.dataset_to_info[key]['meta'])
        return results

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        result = self.dataset_to_info[dataset_id]['meta']
        result['id'] = dataset_id
        return result

    def category_names(self, email, dataset_id):
        results = []
        json_data = self.dataset_to_info[dataset_id]['json_data']
        categories = json_data['categories']
        for category_name in categories:
            category = categories[category_name]
            for category_key in category:
                r = dict(category=category_name, original=category_key, new=category[category_key]['new'])
                results.append(r)
        return results

    def upsert_category_name(self, email, category, dataset_id, original_name, new_name):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        category_entity = json_data['categories'].get(category)
        if category_entity is None:
            category_entity = {}
            json_data['categories'][category] = category_entity
        if new_name == '':
            if original_name in category_entity:
                del category_entity[original_name]
        else:
            entity = dict(new=new_name)
            category_entity[original_name] = entity
            if dataset_id is not None:
                entity['dataset_id'] = dataset_id
            if email is not None:
                entity['email'] = email
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def dataset_filters(self, email, dataset_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        results = []
        filters = json_data['filters']
        for key in filters:
            r = filters[key]
            r['id'] = key
            results.append(r)
        return results

    def get_feature_sets(self, email, dataset_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        return json_data.get('markers', [])

    def delete_feature_set(self, email, dataset_id, set_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        markers = json_data['markers']
        for i in range(len(markers)):
            if markers[i]['id'] == set_id:
                markers.pop(i)
                break
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        if set_id is None:
            set_id = unique_id()
        else:
            self.delete_feature_set(email=email, dataset_id=dataset_id, set_id=set_id)
        json_data = self.dataset_to_info[dataset_id]['json_data']
        markers = json_data.get('markers')
        if markers is None:
            markers = []
            json_data['markers'] = markers
        markers.append(dict(id=set_id, features=features, name=name, category=category))
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return set_id

    def delete_dataset_filter(self, email, dataset_id, filter_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        del json_data['filters'][filter_id]
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def get_dataset_filter(self, email, dataset_id, filter_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        return json_data['filters'][filter_id]

    def upsert_dataset_filter(self, email, dataset_id, filter_id, filter_name, filter_notes, dataset_filter):
        if filter_id is None:
            filter_id = unique_id()
        json_data = self.dataset_to_info[dataset_id]['json_data']
        entity = json_data['filters'].get(filter_id)
        if entity is None:
            entity = {}
            json_data['filters'][filter_id] = entity
        if filter_name is not None:
            entity['name'] = filter_name
        if dataset_filter is not None:
            entity['value'] = json.dumps(dataset_filter)
        if email is not None:
            entity['email'] = email
        if dataset_id is not None:
            entity['dataset_id'] = dataset_id
        if filter_notes is not None:
            entity['notes'] = filter_notes
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return filter_id

    def create_job(self, email, dataset_id, job_name, job_type, params):
        job_id = unique_id()
        self.job_id_to_job[job_id] = dict(id=job_id, dataset_id=dataset_id, name=job_name, type=job_type, params=params,
            status=None, result=None, submitted=datetime.datetime.utcnow())
        return job_id

    def get_job(self, email, job_id, return_result):
        job = self.job_id_to_job[job_id]
        if return_result:
            return job['result']
        return dict(id=job['id'], name=job['name'], type=job['type'], status=job['status'], submitted=job['submitted'])

    def get_jobs(self, email, dataset_id):
        results = []
        for job in self.job_id_to_job.values():
            results.append(dict(id=job['id'], name=job['name'], type=job['type'], status=job['status'],
                submitted=job['submitted']))
        return results

    def delete_job(self, email, job_id):
        del self.job_id_to_job[job_id]

    def update_job(self, email, job_id, status, result):
        job = self.job_id_to_job[job_id]
        job['status'] = status
        job['result'] = result



import json
import os

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import scipy.sparse

from cirrocumulus.abstract_dataset import AbstractDataset


class ParquetDataset(AbstractDataset):

    def __init__(self):
        super().__init__()
        self.cached_dataset_id = None
        self.cached_data = {}

    def get_suffixes(self):
        return ['parquet', 'pq']

    def read_summarized(self, file_system, path, obs_keys=[], var_keys=[], index=False, rename=False, dataset=None):
        result_df = pd.DataFrame()
        if index:
            df = pq.read_table(path + '/index.parquet', filesystem=file_system).to_pandas()

            for column in df:
                result_df[column] = df[column]

        for key in var_keys + obs_keys:
            df = pq.read_table(path + '/' + key + '.parquet', filesystem=file_system).to_pandas()
            if rename:
                for column in df:
                    result_df['{}_{}'.format(key, column)] = df[column]
            else:
                for column in df:
                    result_df[column] = df[column]
        return result_df

    def read_data_sparse(self, file_system, path, keys, dataset=None, schema=None):
        dataset_id = dataset['id']
        # if basis, read index to get bins, x, and y
        result_df = None
        var_keys = keys.pop('X', [])
        obs_keys = keys.pop('obs', [])
        basis = keys.pop('basis', [])

        if len(var_keys) > 0:
            data = []
            row = []
            col = []
            node_path = os.path.join(path, 'X')
            shape = schema['shape']
            for i in range(len(var_keys)):
                key = var_keys[i]
                df = pq.read_table(node_path + '/' + key + '.parquet', filesystem=file_system).to_pandas()
                data.append(df['value'])
                row.append(df['index'])
                col.append(np.repeat(i, len(df)))

            data = np.concatenate(data)
            row = np.concatenate(row)
            col = np.concatenate(col)
            # X = scipy.sparse.csr_matrix((data, (row, col)), shape=(shape[0], len(var_keys)))
            X = scipy.sparse.csc_matrix((data, (row, col)), shape=(shape[0], len(var_keys)))
            result_df = pd.DataFrame.sparse.from_spmatrix(X, columns=var_keys)

        for node in keys.keys():
            if result_df is None:
                result_df = pd.DataFrame()
            features = keys[node]
            node_path = os.path.join(path, node)
            for i in range(len(features)):
                key = features[i]
                df = pq.read_table(node_path + '/' + key + '.parquet', filesystem=file_system).to_pandas()
                result_df[key] = df['value']
        if len(obs_keys) > 0:
            if result_df is None:
                result_df = pd.DataFrame()
            for key in obs_keys:
                node_path = os.path.join(path, 'obs')
                cache_key = str(dataset_id) + '-' + key
                cached_value = self.cached_data.get(cache_key)
                if cached_value is None:
                    df = pq.read_table(node_path + '/' + key + '.parquet', filesystem=file_system,
                        columns=['value']).to_pandas()
                    # ignore index in obs for now
                    cached_value = df['value']
                result_df[key] = cached_value
                self.cached_data[cache_key] = cached_value

        if basis is not None and len(basis) > 0:
            if result_df is None:
                result_df = pd.DataFrame()
            obsm_path = os.path.join(path, 'obsm')
            for b in basis:
                cache_key = str(dataset_id) + '-' + b['full_name']
                is_precomputed = b['precomputed']
                if is_precomputed:
                    columns_to_fetch = [b['full_name']]
                else:  # need coordinates
                    columns_to_fetch = b['coordinate_columns']
                cached_value = self.cached_data.get(cache_key)
                if cached_value is None:
                    basis_path = (obsm_path + '/') + (b['full_name' if is_precomputed else 'name']) + '.parquet'
                    df = pq.read_table(basis_path, filesystem=file_system, columns=columns_to_fetch).to_pandas()
                    cached_value = df
                    self.cached_data[cache_key] = cached_value

                for c in columns_to_fetch:
                    result_df[c] = cached_value[c]
        if result_df is None:
            shape = schema['shape']
            result_df = pd.DataFrame(index=pd.RangeIndex(shape[0]))
        return result_df

    def read_data_dense(self, file_system, path, keys=None, dataset=None, schema=None):
        # dense parquet file
        var_keys = keys.pop('X', [])
        obs_keys = keys.pop('obs', [])
        basis = keys.pop('basis', [])
        all_keys = obs_keys + var_keys
        for key in keys.keys():
            all_keys += keys[key]

        if basis is not None and len(basis) > 0:
            for b in basis:
                all_keys += b['coordinate_columns']
        if len(all_keys) == 0:
            return pq.read_table(path, filesystem=file_system, columns=['index']).to_pandas()
        return pq.read_table(path, filesystem=file_system, columns=all_keys).to_pandas()

    def read_dataset(self, file_system, path, keys=None, dataset=None, schema=None):
        dataset_id = dataset['id']
        if self.cached_dataset_id != dataset_id:
            self.cached_dataset_id = dataset_id
            self.cached_data = {}
        if keys is None:
            keys = {}
        # path is directory
        keys = keys.copy()
        if not path.endswith('.parquet'):
            return self.read_data_sparse(file_system, path, keys, dataset, schema)

        return self.read_data_dense(file_system, path, keys, dataset, schema)

    def schema(self, file_system, path):
        if path.endswith('.json') or path.endswith('.json.gz'):
            return super().schema(file_system, path)
        elif path.endswith('.parquet'):
            metadata = None
            with file_system.open(path, 'rb') as s:
                parquet_file = pq.ParquetFile(s)
                schema = parquet_file.schema.to_arrow_schema()
                result = {'version': '1'}
                for metadata_key in [b'pegasus']:
                    if metadata_key in schema.metadata:
                        metadata = json.loads(schema.metadata[metadata_key])
                        break
                obs = []
                obs_cat = []
                if metadata is None:
                    # obs, the obsm, then var
                    names = parquet_file.schema.names
                    section = 'obs'
                    result['embeddings'] = []
                    result['var'] = []
                    embedding_basename_to_count = {}
                    for name in names:
                        if section == 'obs' and name.startswith('X_'):
                            section = 'obsm'
                        if section == 'obs':
                            field = schema.field(name)
                            if isinstance(field.type, pa.lib.DictionaryType):
                                obs_cat.append(name)
                            else:
                                obs.append(name)
                        if section == 'obsm' and not name.startswith('X_'):
                            section = 'var'
                        if section == 'obsm':  # X_fitsne_1', 'X_fitsne_2, etc
                            basename = name[0:name.rindex('_')]
                            count = embedding_basename_to_count.get(basename, 0)
                            embedding_basename_to_count[basename] = count + 1
                        if section == 'var':
                            result['var'].append(name)
                    for name in embedding_basename_to_count.keys():
                        result['embeddings'].append(dict(name=name, dimensions=2))
                        if embedding_basename_to_count[name] == 3:
                            result['embeddings'].append(dict(name=name, dimensions=3))
                else:
                    all_obs = metadata['obs']

                    for name in all_obs:
                        field = schema.field(name)
                        if isinstance(field.type, pa.lib.DictionaryType):
                            obs_cat.append(name)
                        else:
                            obs.append(name)
                    result['var'] = metadata['var']
                    result['embeddings'] = metadata['obsm']

                result['obs'] = obs
                result['obsCat'] = obs_cat
                result['shape'] = (parquet_file.metadata.num_rows, len(result['var']))
                return result
        else:  # directory
            return super().schema(file_system, os.path.join(path, 'index.json.gz'))



import pandas as pd
from pytest import approx

from cirrocumulus.data_processing import handle_data


def diff_measures(test_data, summary, fields, is_x):
    X = test_data[:, fields].X if is_x else test_data.obs[fields].values
    my_mean = X.mean(axis=0)
    my_min = X.min(axis=0)
    my_max = X.max(axis=0)
    my_sum = X.sum(axis=0)
    my_expressed = (X != 0).sum(axis=0)

    for i in range(len(fields)):
        field = fields[i]
        field_summary = summary[field]
        assert my_min[i] == field_summary['min'], 'min'
        assert my_max[i] == field_summary['max'], 'max'
        assert my_sum[i] == approx(field_summary['sum'], abs=0.001), 'sum'
        assert my_mean[i] == approx(field_summary['mean'], abs=0.0000001), 'mean'

        if 'numExpressed' in field_summary:
            assert my_expressed[i] == field_summary['numExpressed'], 'numExpressed'


def diff_dimensions(test_data, summary, fields):
    for key in fields:
        value_counts = test_data.obs[key].value_counts()
        dimension_summary = summary[key]
        df = pd.DataFrame.from_dict(dimension_summary)
        df['categories'] = df['categories'].astype('category')
        df = df.set_index('categories')
        df = df.iloc[df.index.get_indexer_for(value_counts.index)]
        assert (df['counts'] - value_counts).sum() == 0, key


def test_measure(dataset_api, input_dataset, test_data, measures):
    process_results = handle_data(dataset_api=dataset_api, dataset=input_dataset, stats=dict(measures=measures))
    summary = process_results['summary']
    diff_measures(test_data, summary, measures, True)


def test_continuous_obs(dataset_api, input_dataset, test_data, continuous_obs):
    process_results = handle_data(dataset_api=dataset_api, dataset=input_dataset,
        stats=dict(measures=list(map(lambda x: 'obs/' + x, continuous_obs))))
    summary = process_results['summary']
    diff_measures(test_data, summary, continuous_obs, False)


def test_dimension(dataset_api, input_dataset, test_data, dimensions):
    process_results = handle_data(dataset_api=dataset_api, dataset=input_dataset,
        stats=dict(dimensions=dimensions))
    summary = process_results['summary']
    diff_dimensions(test_data, summary, dimensions)


def test_all(dataset_api, input_dataset, test_data, measures, dimensions, continuous_obs):
    process_results = handle_data(dataset_api=dataset_api, dataset=input_dataset,
        stats=dict(measures=measures + list(map(lambda x: 'obs/' + x, continuous_obs)),
            dimensions=dimensions))
    summary = process_results['summary']
    diff_measures(test_data, summary, measures, True)
    diff_measures(test_data, summary, continuous_obs, False)
    diff_dimensions(test_data, summary, dimensions)



import gzip
import json
import logging
import os

import numpy as np
import pandas as pd
import pandas._libs.json as ujson
import scipy.sparse

logger = logging.getLogger("cirro")

LINE_END = '\n'.encode('UTF-8')


def write_jsonl(d, f, name, index, compress=False):
    output = {}
    output[name] = d
    c = ujson.dumps(output, double_precision=2, orient='values').encode('UTF-8')
    if compress:
        c = gzip.compress(c)
    start = f.tell()
    end = start + len(c)
    index[name] = [start, end - 1]
    f.write(c)
    f.write(LINE_END)


def read_adata_jsonl(path, keys):
    with open(path + '.idx.json', 'rt') as f:
        index = json.load(f)
    compress = False

    def read_key(f, index_key):
        start, end = index['index'][index_key]
        f.seek(start)
        b = f.read(end - start + 1)
        if compress:
            b = gzip.decompress(b)
        return json.loads(b.decode('UTF-8'))

    df = pd.DataFrame()
    with open(path, 'rb') as f:
        shape = read_key('schema')['shape']
        for key in keys:
            value = read_key(f, key)
            if isinstance(value, dict):
                if 'index' in value:
                    array = np.zeros(shape[0])
                    array[value['index']] = value['value']
                    df[key] = pd.arrays.SparseArray(array)
                else:
                    # obsm
                    for obsm_index in value:
                        df[obsm_index] = value[obsm_index]
            else:
                df[key] = value

    return df


def save_adata_jsonl(adata, schema, output_path):
    logger.info('Save adata')
    compress = False
    index = {}  # key to byte start-end

    if not output_path.lower().endswith('.jsonl'):
        output_path += '.jsonl'

    with open(output_path, 'wb') as f:
        save_adata_X(adata, f, index, compress)
        save_data_obs(adata, f, index, compress)
        save_data_obsm(adata, f, index, compress)
        write_jsonl(schema, f, 'schema', index)

    with open(output_path + '.idx.json', 'wt') as f:
        # json.dump(result, f)

        result = dict(index=index, file=os.path.basename(output_path))
        f.write(ujson.dumps(result, double_precision=2, orient='values'))


def save_adata_X(adata, f, index, compress):
    adata_X = adata.X
    names = adata.var.index
    is_sparse = scipy.sparse.issparse(adata_X)
    if is_sparse and scipy.sparse.isspmatrix_csr(adata_X):
        adata_X = adata_X.tocsc()

    for j in range(adata_X.shape[1]):
        X = adata_X[:, j]
        if is_sparse:
            X = X.toarray().flatten()
        indices = np.where(X != 0)[0]
        values = X[indices]
        write_jsonl(dict(index=indices, value=values), f, names[j], index, compress)
        if j > 0 and (j + 1) % 1000 == 0:
            logger.info('Wrote adata X {}/{}'.format(j + 1, adata_X.shape[1]))


def save_data_obsm(adata, f, index, compress):
    logger.info('writing adata obsm')

    for name in adata.obsm.keys():
        m = adata.obsm[name]
        dimensions = 3 if m.shape[1] > 2 else 2
        d = {}
        for i in range(dimensions):
            d[name + '_' + str(i + 1)] = m[:, i].astype('float32')
        write_jsonl(d, f, name, index, compress)


def save_data_obs(adata, f, index, compress):
    logger.info('writing adata obs')
    for name in adata.obs:
        value = adata.obs[name]
        write_jsonl(value, f, name, index, compress)
    write_jsonl(adata.obs.index.values, f, 'index', index, compress)
